"""Base class for all offax_psfs.fits files."""

from multiprocessing import Pool
from pathlib import Path

import astropy.io.fits as pyfits
import astropy.units as u
import numpy as np
from astropy.units import Quantity
from lod_unit import lod
from numpy.typing import NDArray

from yippy.util import convert_to_lod, create_shift_mask, fft_shift

from .logger import logger


class OffAx:
    """Class for handling off-axis PSFs in pure Python.

    This class loads and processes PSF data from the yield input package (YIP).
    It currently supports oneD and quater symmetric PSF YIPs. The primary use
    is to interpolate the PSF data to a given x/y position. This is done by
    calling the OffAx object with the x/y position as arguments, which itself
    calls the psf object after converting units.

    Attributes:
        pixel_scale (Quantity):
            Pixel scale of the PSF data in lambda/D.
        center_x (Quantity):
            Central x position in the PSF data.
        center_y (Quantity):
            Central y position in the PSF data.
        psf:
            Instance of the appropriate PSF class (e.g., OneD, TwoD) based on input YIP.

    Args:
        yip_dir (Path):
            Path to the directory containing PSF and offset data.
        logger (Logger):
            Logger for logging events and information.
        offax_data_file (str):
            Name of the file containing the PSF data.
        offax_offsets_file (str):
            Name of the file containing the offsets data.
        pixel_scale (Quantity):
            Pixel scale of the PSF data in lambda/D.
    """

    def __init__(
        self,
        yip_dir: Path,
        offax_data_file: str,
        offax_offsets_file: str,
        pixel_scale: Quantity,
        x_symmetric: bool,
        y_symmetric: bool,
    ) -> None:
        """Initializes the OffAx class by loading PSF and offset data from YIP."""
        # Pixel scale in lambda/D
        self.pixel_scale = pixel_scale

        # Load symmetry
        self.x_symmetric = x_symmetric
        self.y_symmetric = y_symmetric

        # Load off-axis PSF data (e.g. the planet) (unitless intensity maps)
        psfs = pyfits.getdata(Path(yip_dir, offax_data_file), 0)

        # Save the center of the pixel array, which is used for converting to
        # lambda/D when the x/y positions are in pixels.
        self.center_x = psfs.shape[1] / 2 * u.pix
        self.center_y = psfs.shape[2] / 2 * u.pix

        # Load the offset list, which is in units of lambda/D
        offsets = pyfits.getdata(Path(yip_dir, offax_offsets_file), 0)

        # Check whether offsets are given as 1D or 2D
        one_d_offsets = len(offsets.shape) == 1
        if one_d_offsets:
            # Add a second dimension if the offsets are 1D
            offsets = np.vstack((offsets, np.zeros_like(offsets)))

        if len(offsets.shape) > 1:
            if (offsets.shape[1] != 2) and (offsets.shape[0] == 2):
                # This condition occurs when the offsets is transposed
                # from the expected format
                offsets = offsets.T
        assert len(offsets) == psfs.shape[0], (
            "Offsets and PSFs do not have the same number of elements"
        )

        ########################################################################
        # Determine the format of the input coronagraph files so we can handle #
        # the coronagraph correctly (e.g. radially symmetric in x direction)   #
        ########################################################################

        # Get the unique values of the offset list so that we can format the
        # data into
        offsets_x = np.unique(offsets[:, 0])
        offsets_y = np.unique(offsets[:, 1])

        if len(offsets_x) == 1:
            logger.info(f"{yip_dir.stem} is radially symmetric")
            self.type = "1d"
            # Instead of handling angles for 1dy, swap the x and y
            offsets_x, offsets_y = (offsets_y, offsets_x)
            offsets = np.vstack((offsets_y, offsets_x)).T
            raise NotImplementedError(
                (
                    "Verify that the PSFs are correct for this case!"
                    " I don't have a test file for this case yet but I think they"
                    " probably need to be rotated by 90 degrees."
                )
            )
        elif len(offsets_y) == 1:
            logger.info(f"{yip_dir.stem} is radially symmetric")
            self.type = "1d"
        elif np.min(offsets) >= 0:
            logger.info(f"{yip_dir.stem} is quarterly symmetric")
            self.type = "2dq"
            # Check if 0 is included
            if 0 not in offsets_x:
                # Need to mirror the PSFs across the x axis for the interpolation
                min_x = np.min(offsets_x)
                # Add the mirrored offset to the original offsets
                offsets_x = np.insert(offsets_x, 0, -min_x)

                # Get all the PSFs that are at the minimum x value
                min_x_psfs = psfs[offsets[:, 0] == min_x]

                # Add the mirrored PSFs to the original PSFs
                psfs = np.insert(psfs, 0, np.flip(min_x_psfs, axis=2), axis=0)

                # Add the mirrored offset to the original offsets
                new_offsets = np.array(np.meshgrid(-min_x, offsets_y)).T.reshape(-1, 2)

                # Create an array of negative_min_x with the y offsets
                offsets = np.insert(offsets, 0, new_offsets, axis=0)
            if 0 not in offsets_y:
                # Need to mirror the PSFs across the y axis for the interpolation
                min_y = np.min(offsets_y)
                offsets_y = np.insert(offsets_y, 0, -min_y)

                # Get all the PSFs that are at the minimum y value
                min_y_psfs = psfs[offsets[:, 1] == min_y]

                # Add the mirrored PSFs to the original PSFs
                psfs = np.insert(psfs, 0, np.flip(min_y_psfs, axis=1), axis=0)

                # Add the mirrored offset to the original offsets
                new_offsets = np.array(np.meshgrid(offsets_x, -min_y)).T.reshape(-1, 2)

                # Create an array of negative_min_x with the y offsets
                offsets = np.insert(offsets, 0, new_offsets, axis=0)
        else:
            logger.info(f"{yip_dir.stem} response is full 2D")
            self.type = "2df"

        # Initialize the reshaped PSFs array to allow us to index by the offsets
        self.reshaped_psfs = np.empty((len(offsets_x), len(offsets_y), *psfs.shape[1:]))
        self.x_inds = np.searchsorted(offsets_x, offsets[:, 0])
        self.y_inds = np.searchsorted(offsets_y, offsets[:, 1])
        self.reshaped_psfs[self.x_inds, self.y_inds] = psfs
        self.n_psfs = len(offsets)

        self.x_offsets = offsets_x
        self.y_offsets = offsets_y
        # self.offsets = np.array(list(product(self.x_offsets, self.y_offsets)))
        self.x_range = np.array([self.x_offsets[0], self.x_offsets[-1]])
        self.y_range = np.array([self.y_offsets[0], self.y_offsets[-1]])

    def create_psf(self, x: float, y: float):
        """Creates and returns the PSF at the specified off-axis position.

        Interpolates and returns the Point Spread Function (PSF) at the specified
        off-axis position (x, y). If the exact (x, y) position matches one of the
        PSFs in the YIP, that PSF is returned directly. Otherwise, the PSFs
        from the surrounding positions are combined using Gaussian weighting and
        Fourier interpolation to produce an interpolated PSF.

        Args:
            x (float):
                The x-coordinate of the off-axis position.
            y (float):
                The y-coordinate of the off-axis position.

        Returns:
            np.ndarray:
                The interpolated PSF corresponding to the input (x, y) position.

        Notes:
            - If `self.type` is "1d", the (x, y) position is converted to a
            radial separation and angle for interpolation.
            - If `self.type` is "2dq", the (x, y) position is mirrored to the
            first quadrant, and the PSF is flipped accordingly after
            interpolation.
            - Gaussian weighting is used to combine the nearest PSFs when the
            exact (x, y) position does not match any precomputed PSF. The
            weighting is based on the distance from the input position.
            - The PSFs are shifted to align with the input position before
            combining, and the final PSF is normalized by the cumulative weight
            for each pixel.

        """
        # Set default values
        flip_lr, flip_ud = False, False

        # Check for exact matches
        if x in self.x_offsets and y in self.y_offsets:
            x_ind = np.searchsorted(self.x_offsets, x)
            y_ind = np.searchsorted(self.y_offsets, y)
            return self.reshaped_psfs[x_ind, y_ind]

        # Translate position based on type
        if self.type == "1d":
            flip_lr, flip_ud = x < 0, y < 0
            sep = np.sqrt(x**2 + y**2)
            _x, _y = sep, 0
        elif self.type == "2dq":
            flip_lr, flip_ud = x < 0, y < 0
            _x, _y = abs(x), abs(y)
        else:
            _x, _y = x, y

        # Get indices of nearest PSFs, in x and y directions
        x_match = _x in self.x_offsets
        _x_search = np.searchsorted(self.x_offsets, _x)
        if x_match:
            # If the x value is an exact match, we only need one index
            x_inds = _x_search.reshape(-1)
        else:
            x_inds = np.array([_x_search - 1, _x_search])

        y_match = _y in self.y_offsets
        _y_search = np.searchsorted(self.y_offsets, _y)
        if y_match:
            # If the y value is an exact match, we only need one index
            y_inds = _y_search.reshape(-1)
        else:
            y_inds = np.array([_y_search - 1, _y_search])

        x_vals, y_vals = self.x_offsets[x_inds], self.y_offsets[y_inds]
        # Get the indices of the nearest PSFs to the input (x, y)
        near_inds = np.array(np.meshgrid(x_inds, y_inds)).T.reshape(-1, 2)

        # Get the (x, y) offsets of the nearest PSFs to the input (x, y)
        near_offsets = np.array(np.meshgrid(x_vals, y_vals)).T.reshape(-1, 2)

        # Get the PSFs at the nearest offsets
        near_psfs = self.reshaped_psfs[near_inds[:, 0], near_inds[:, 1]]

        # Combine the PSFs
        if len(near_psfs) > 1:
            # Get the shift (in pixels) required to align with the input (x, y)
            near_shifts = (np.array([_x, _y]) - near_offsets) / self.pixel_scale.value

            # Calculate the distance of each PSF from the input (x, y)
            near_diffs = np.linalg.norm(near_shifts, axis=1)

            # Gaussian weighting
            sigma = 0.25
            weights = np.exp(-(near_diffs**2) / (2 * sigma**2))

            # Normalize the weights
            weights /= weights.sum()

            # Initialize the PSF array
            psf = np.zeros_like(near_psfs[0])

            # Initialize the weight array
            # This weight system is used because shifting a PSF right by one pixel
            # will leave a blank pixel on the left side of the image. The weight
            # array keeps track of which PSFs have contributions for each pixel.
            weight_array = np.zeros_like(psf)
            for i, near_psf in enumerate(near_psfs):
                shifted_psf = fft_shift(near_psf, *near_shifts[i])
                weight_mask = create_shift_mask(near_psf, *near_shifts[i], weights[i])
                # Add the weighted PSF to the total PSF
                psf += weight_mask * shifted_psf
                # Keep track of the weight for each pixel
                weight_array += weight_mask
            # Divide each pixel by its weight to get the final PSF
            psf /= weight_array
        else:
            psf = near_psfs[0]

        # Apply any necessary flips before shifting
        if self.x_symmetric and flip_lr:
            psf = np.fliplr(psf)
            remaining_x_shift = x + _x
        else:
            remaining_x_shift = x - _x
        if self.y_symmetric and flip_ud:
            psf = np.flipud(psf)
            remaining_y_shift = y + _y
        else:
            remaining_y_shift = y - _y

        if remaining_x_shift != 0 or remaining_y_shift != 0:
            psf = fft_shift(
                psf,
                remaining_x_shift / self.pixel_scale.value,
                remaining_y_shift / self.pixel_scale.value,
            )

        return psf

    def create_psfs(self, x: NDArray, y: NDArray) -> NDArray:
        """Creates and returns the PSFs at the specified off-axis positions."""
        psfs = np.empty((len(x), *self.reshaped_psfs.shape[2:]))
        for i in range(len(x)):
            psfs[i] = self.create_psf(x[i], y[i])
        return psfs

    def create_psfs_parallel(
        self,
        x: np.ndarray,
        y: np.ndarray,
        lam=None,
        D=None,
        dist=None,
        workers: int = 4,
    ) -> np.ndarray:
        """Compute PSFs for batches of (x, y) arrays using multiprocessing.

        Args:
            x (np.ndarray):
                Array of x positions.
            y (np.ndarray):
                Array of y positions.
            lam (astropy.units.Quantity):
                Wavelength of the observation
            D (astropy.units.Quantity):
                Diameter of the telescope
            dist (astropy.units.Quantity):
                Distance to the system
            workers (int):
                Number of parallel processes to use.

        Returns:
            np.ndarray:
                A stacked numpy array of the computed PSFs with shape (N,
                height, width), where N = len(x).
        """
        if isinstance(x, Quantity):
            # Convert the x and y positions to lambda/D if they are in pixels
            if x.unit != lod:
                x = convert_to_lod(x, self.center_x, self.pixel_scale, lam, D, dist)
            else:
                x = x.value
        if isinstance(y, Quantity):
            if y.unit != lod:
                y = convert_to_lod(y, self.center_y, self.pixel_scale, lam, D, dist)
            else:
                y = y.value
        # For each x[i], we create a column of identical x-values and the full y-array.
        # This way, each process will handle (len(y)) points for that particular x[i].
        args = []
        for xi in x:
            # Create an array filled with xi, same shape as y
            x_col = np.full_like(y, xi)
            args.append((x_col, y))

        # Each process will handle a single (x_col, y) pair and run self.create_psfs
        # which returns a 3D stack: shape = (len(y), height, width).
        with Pool(processes=workers) as pool:
            psf_rows = pool.starmap(self.create_psfs, args)

        # psf_rows is now a list of arrays, each of shape (len(y), height, width)
        # Stack them along a new axis to form (len(x), len(y), height, width)
        psf_datacube = np.stack(psf_rows, axis=0)
        return psf_datacube

    def __call__(
        self, x: Quantity, y: Quantity, lam=None, D=None, dist=None
    ) -> NDArray:
        """Return the PSF at the given x/y position.

        This function (via util.convert_to_lod) has the following assumptions
        on the x/y values provided:
            - If units are pixels, they follow the 00LL convention. As in the
              (0,0) point is the lower left corner of the image.
            - If the x/y values are in lambda/D, angular, or length units the
              (0,0) point is the center of the image, where the star is
              (hopefully) located.

        Args:
            x (astropy.units.Quantity):
                x position. Can be either units of pixel, lod, an angular
                unit (e.g. arcsec), or a length unit (e.g. AU)
            y (astropy.units.Quantity):
                y position. Can be either units of pixel, lod, an angular
                unit (e.g. arcsec), or a length unit (e.g. AU)
            lam (astropy.units.Quantity):
                Wavelength of the observation
            D (astropy.units.Quantity):
                Diameter of the telescope
            dist (astropy.units.Quantity):
                Distance to the system

        Returns:
            NDArray:
                The PSF at the given x/y position
        """
        if isinstance(x, Quantity):
            # Convert the x and y positions to lambda/D if they are in pixels
            if x.unit != lod:
                x = convert_to_lod(x, self.center_x, self.pixel_scale, lam, D, dist)
            x = x.value
        if isinstance(y, Quantity):
            if y.unit != lod:
                y = convert_to_lod(y, self.center_y, self.pixel_scale, lam, D, dist)
            y = y.value

        if np.isscalar(x) and np.isscalar(y):
            return self.create_psf(x, y)
        else:
            return self.create_psfs(x, y)



"""Module for handling stellar intensity data from stellar_intens files."""

from pathlib import Path

import astropy.io.fits as pyfits
import numpy as np
from astropy.units import Quantity
from lod_unit import lod
from scipy.interpolate import CubicSpline

from .util import convert_to_lod


class StellarIntens:
    """Class to handle and interpolate stellar intensity data.

    This class loads stellar intensity data and corresponding stellar diameters,
    providing an interpolation interface to get the stellar intensity map for a given
    stellar diameter.

    Attributes:
        ln_interp (CubicSpline):
            A spline interpolator that operates on the natural logarithm of the
            stellar intensities to ensure that interpolated values are non-negative.

    Args:
        yip_dir (Path):
            Path to the directory containing the yield input package (YIP).
        stellar_intens_file (str):
            Filename of the FITS file containing the stellar intensity data.
        stellar_diam_file (str):
            Filename of the FITS file containing the stellar diameters.
    """

    def __init__(
        self,
        yip_dir: Path,
        stellar_intens_file: str,
        stellar_diam_file: str,
    ) -> None:
        """Initializes StellarIntens class by loading data and creating the interpolant.

        Stellar intensity data and stellar diameters are loaded from FITS files,
        and a natural logarithm-based cubic spline interpolator is set up to
        facilitate interpolation.

        Args:
            yip_dir (Path):
                The directory where the stellar intensity and diameter FITS
                files are located.
            stellar_intens_file (str):
                The filename of the FITS file containing unitless arrays of stellar
                intensities.
            stellar_diam_file (str):
                The filename of the FITS file containing arrays of stellar diameters in
                lambda/D units.
        """
        # Load on-axis stellar intensity PSF data
        psfs = pyfits.getdata(Path(yip_dir, stellar_intens_file), 0)

        # Load the stellar angular diameters in units of lambda/D
        diams = pyfits.getdata(Path(yip_dir, stellar_diam_file), 0) * lod

        # Interpolate stellar data in logarithmic space to ensure non-negative
        # interpolated values
        self.ln_interp = CubicSpline(diams, np.log(psfs))

    def __call__(self, stellar_diam: Quantity, lam=None, D=None):
        """Returns the stellar intensity map at a specified stellar diameter.

        Stellar intensity is interpolated based on a specified diameter using the
        previously configured cubic spline interpolator. The interpolation considers
        the logarithm of the intensity to ensure that all computed values are positive.

        Args:
            stellar_diam (Quantity):
                The desired stellar diameter for which to retrieve the intensity map, in
                units of lambda/D.
            lam (Quantity, optional):
                Wavelength of observation, required if stellar_diam is in angular units.
            D (Quantity, optional):
                Diameter of the telescope, required if stellar_diam is in angular units.

        Returns:
            NDArray:
                An interpolated 2D map of the stellar intensity at the given
                stellar diameter.
        """
        # Set up conversion from angular units to lambda/D
        if stellar_diam.unit != lod:
            stellar_diam = convert_to_lod(stellar_diam, lam=lam, D=D)

        # Return the exponentiated result of the interpolation to get the
        # actual intensity map
        return np.exp(self.ln_interp(stellar_diam))



"""JAX functions for image processing operations."""

import os
import re
from functools import partial

import jax
import jax.numpy as jnp
from jax import lax


def enable_x64(use_x64=True):
    """Changes the default array type to use 64 bit precision as in NumPy.

    Lovingly borrowed from the numpyro library until JAX provides a more
    convenient way.

    Args:
        use_x64 (Bool):
            when `True`, JAX arrays will use 64 bits by default; else 32 bits.
    """
    if not use_x64:
        use_x64 = bool(os.getenv("JAX_ENABLE_X64", 0))
    jax.config.update("jax_enable_x64", use_x64)


def set_platform(platform=None):
    """Changes platform to CPU, GPU, or TPU.

    This utility only takes effect at the beginning of your program.

    Lovingly borrowed from the numpyro library until JAX provides a more
    convenient way.

    Args:
        platform (str):
           either 'cpu', 'gpu', or 'tpu'.
    """
    if platform is None:
        platform = os.getenv("JAX_PLATFORM_NAME", "cpu")
    jax.config.update("jax_platform_name", platform)


def set_host_device_count(n: int) -> None:
    """Set the number of CPU cores available to XLA.

    By default, XLA considers all CPU cores as one device. This utility tells XLA
    that there are `n` host (CPU) devices available to use. As a consequence, this
    allows parallel mapping in JAX :func:`jax.pmap` to work in CPU platform.

    Lovingly borrowed from the numpyro library until JAX provides a more
    convenient way.

    .. note:: This utility only takes effect at the beginning of your program.
        Under the hood, this sets the environment variable
        `XLA_FLAGS=--xla_force_host_platform_device_count=[num_devices]`, where
        `[num_device]` is the desired number of CPU devices `n`.

    .. warning:: Our understanding of the side effects of using the
        `xla_force_host_platform_device_count` flag in XLA is incomplete. If you
        observe some strange phenomenon when using this utility, please let us
        know through our issue or forum page. More information is available in this
        `JAX issue <https://github.com/google/jax/issues/1408>`_.


    Args:
        n (int):
            number of CPU devices to use.
    """
    xla_flags_str = os.getenv("XLA_FLAGS", "")
    xla_flags = re.sub(
        r"--xla_force_host_platform_device_count=\S+", "", xla_flags_str
    ).split()
    os.environ["XLA_FLAGS"] = " ".join(
        ["--xla_force_host_platform_device_count={}".format(n)] + xla_flags
    )


def get_pad_info(image, pad_factor):
    """Get the padding information for an image.

    Args:
        image (jax.numpy.ndarray):
            The input image to be shifted.
        pad_factor (float):
            The factor by which to pad the image.

    Returns:
        int:
            The number of pixels in the original image.
        int:
            The number of pixels to pad the image.
        int:
            The edge of the image after padding.
        int:
            The number of pixels in the final image.
    """
    n_pixels_orig = image.shape[0]
    n_pad = int(pad_factor * n_pixels_orig)
    img_edge = n_pad + n_pixels_orig
    n_pixels_final = int(2 * n_pixels_orig * pad_factor + n_pixels_orig)
    return n_pixels_orig, n_pad, img_edge, n_pixels_final


def fft_shift_x(image, shift_pixels, phasor):
    """Apply a Fourier shift to an image along the x axis.

    A traceable version of the `fft_shift_1d` function that applies a Fourier
    shift to an image along the x-axis. This allows the function to be traced
    and compiled by JAX.

    Args:
        image (jax.numpy.ndarray):
            The input 2D image to be shifted.
        shift_pixels (float):
            The number of pixels by which to shift the image along the specified axis.
        phasor (jax.numpy.ndarray):
            Precomputed components for the Fourier shift (exp(-2j * pi * fft_freqs)).

    Returns:
        jax.numpy.ndarray:
            The shifted image after applying the Fourier transform and removing
            the padding.
    """
    n_pixels_orig, n_pad, img_edge, n_pixels_final = get_pad_info(image, 1.5)

    # Pad the image with zeros
    padded = lax.pad(image, 0.0, [(n_pad, n_pad, 0), (n_pad, n_pad, 0)])

    # Take the 1D Fourier transform along the specified axis
    padded = jnp.fft.fft(padded, axis=1)

    # Tile the phasor to match the dimensions of the padded image
    # Horizontal shift (x-axis)
    phasor = jnp.tile(phasor**shift_pixels, (padded.shape[0], 1))

    # Apply the phasor along the specified axis
    padded = padded * phasor

    # Reconstruct the image using the inverse Fourier transform along the specified axis
    padded = jnp.real(jnp.fft.ifft(padded, axis=1))

    # Unpad the image to return to the original size
    image = padded[n_pad:img_edge, n_pad:img_edge]

    # Cut any negative values to zero. This occurs in the region with no
    # information in the original image (e.g. the left pixels when moving
    # a PSF rightwards)
    return jnp.maximum(image, 0.0)


def fft_shift_y(image, shift_pixels, phasor):
    """Apply a Fourier shift to an image along the y axis.

    A traceable version of the `fft_shift_1d` function that applies a Fourier
    shift to an image along the y-axis. This allows the function to be traced
    and compiled by JAX.

    Args:
        image (jax.numpy.ndarray):
            The input 2D image to be shifted.
        shift_pixels (float):
            The number of pixels by which to shift the image along the specified axis.
        phasor (jax.numpy.ndarray):
            Precomputed components for the Fourier shift (exp(-2j * pi * fft_freqs)).

    Returns:
        jax.numpy.ndarray:
            The shifted image after applying the Fourier transform and removing
            the padding.
    """
    n_pixels_orig, n_pad, img_edge, n_pixels_final = get_pad_info(image, 1.5)

    # Pad the image with zeros
    padded = lax.pad(image, 0.0, [(n_pad, n_pad, 0), (n_pad, n_pad, 0)])

    # Take the 1D Fourier transform along the specified axis
    padded = jnp.fft.fft(padded, axis=0)

    # Tile the phasor to match the dimensions of the padded image
    # Vertical shift (y-axis)
    phasor = jnp.tile(phasor**shift_pixels, (padded.shape[1], 1)).T

    # Apply the phasor along the specified axis
    padded = padded * phasor

    # Reconstruct the image using the inverse Fourier transform along the specified axis
    padded = jnp.real(jnp.fft.ifft(padded, axis=0))

    # Unpad the image to return to the original size
    image = padded[n_pad:img_edge, n_pad:img_edge]

    # Cut any negative values to zero. This occurs in the region with no
    # information in the original image (e.g. the left pixels when moving
    # a PSF rightwards)
    return jnp.maximum(image, 0.0)


def handle_shift_x_positive(mask, shift_x, x_grid):
    """Zero out the left side of the mask for positive x shifts."""
    n = jnp.ceil(shift_x).astype(jnp.int32)
    n = jnp.clip(n, 0, mask.shape[1])
    # Set mask to 0 where x_grid < n
    return jnp.where(x_grid < n, 0.0, mask)


def handle_shift_x_negative(mask, shift_x, x_grid):
    """Zero out the right side of the mask for negative x shifts."""
    n = jnp.ceil(-shift_x).astype(jnp.int32)
    n = jnp.clip(n, 0, mask.shape[1])
    # Set mask to 0 where x_grid >= (width - n)
    return jnp.where(x_grid >= (mask.shape[1] - n), 0.0, mask)


def handle_shift_y_positive(mask, shift_y, y_grid):
    """Zero out the top side of the mask for positive y shifts."""
    n = jnp.ceil(shift_y).astype(jnp.int32)
    n = jnp.clip(n, 0, mask.shape[0])
    # Set mask to 0 where y_grid < n
    return jnp.where(y_grid < n, 0.0, mask)


def handle_shift_y_negative(mask, shift_y, y_grid):
    """Zero out the bottom side of the mask for negative y shifts."""
    n = jnp.ceil(-shift_y).astype(jnp.int32)
    n = jnp.clip(n, 0, mask.shape[0])
    # Set mask to 0 where y_grid >= (height - n)
    return jnp.where(y_grid >= (mask.shape[0] - n), 0.0, mask)


def create_shift_mask(psf, shift_x, shift_y, x_grid, y_grid, fill_val=1):
    """Create a mask to identify valid pixels to average.

    This function is useful because when the PSF is shifted there are empty
    pixels, since they were outside the initial image, and should not be
    included in the final average.

    Args:
        psf (jax.numpy.ndarray):
            The PSF image to shift.
        shift_x (float):
            The shift in the x direction.
        shift_y (float):
            The shift in the y direction.
        x_grid (jax.numpy.ndarray):
            The x-coordinate grid.
        y_grid (jax.numpy.ndarray):
            The y-coordinate grid.
        fill_val (float, optional):
            The value to fill the mask with.

    Returns:
        jax.numpy.ndarray:
            The mask to identify valid pixels to average.
    """
    mask = jnp.full_like(psf, fill_val)

    mask = lax.cond(
        jnp.sign(shift_x) == 1,
        # Zero out the left side
        lambda x: handle_shift_x_positive(x, shift_x, x_grid),
        # Zero out the right side
        lambda x: handle_shift_x_negative(x, shift_x, x_grid),
        mask,
    )

    mask = lax.cond(
        jnp.sign(shift_y) == 1,
        # Zero out the bottom side
        lambda x: handle_shift_y_positive(x, shift_y, y_grid),
        # Zero out the top side
        lambda x: handle_shift_y_negative(x, shift_y, y_grid),
        mask,
    )

    return mask


def get_near_inds_offsets_1D(
    x_offsets: jnp.ndarray, y_offsets: jnp.ndarray, _x: float, _y: float
):
    """Computes the nearest indices and offsets for the given _x position in 1D.

    Args:
        x_offsets (jnp.ndarray): 1D array of x offsets, sorted in ascending order.
        y_offsets (jnp.ndarray): 1D array of y offsets.
        _x (float): The x-coordinate for which to find nearby offsets.

    Returns:
        Tuple[jnp.ndarray, jnp.ndarray]:
            - near_inds: Array of shape (2,) containing the indices of the two
              surrounding points.
            - near_offsets: Array of shape (2,) containing the corresponding x offsets.
    """
    # Find insertion index for _x
    _x_ind = jnp.searchsorted(x_offsets, _x, side="left")

    # Handle boundary conditions
    x_ind_low = jnp.clip(_x_ind - 1, 0, x_offsets.size - 1)
    x_ind_high = jnp.clip(_x_ind, 0, x_offsets.size - 1)

    y_ind = 0

    # Collect the two nearest indices
    near_inds = jnp.array([[x_ind_low, y_ind], [x_ind_high, y_ind]])

    # Extract the corresponding x offset values
    x_vals_low = x_offsets[x_ind_low]
    x_vals_high = x_offsets[x_ind_high]
    y_val = y_offsets[y_ind]

    near_offsets = jnp.array([[x_vals_low, y_val], [x_vals_high, y_val]])

    return near_inds, near_offsets


def create_avg_psf_1D(
    x: float,
    y: float,
    pixel_scale: float,
    x_offsets: jnp.ndarray,
    y_offsets: jnp.ndarray,
    x_grid: jnp.ndarray,
    y_grid: jnp.ndarray,
    x_phasor: jnp.ndarray,
    y_phasor: jnp.ndarray,
    reshaped_psfs: jnp.ndarray,
):
    """Creates and returns the PSF at the specified off-axis position using JAX."""
    # The core logic is similar to OffAx, but uses JAX operations
    _x, _y = convert_xy_1D(x, y)

    near_inds, near_offsets = get_near_inds_offsets_1D(x_offsets, y_offsets, _x, _y)

    near_psfs = reshaped_psfs[near_inds[:, 0], near_inds[:, 1]]

    near_shifts = (jnp.array([_x, _y]) - near_offsets) / pixel_scale
    near_diffs = jnp.linalg.norm(near_shifts, axis=1)
    sigma = 0.25
    # Adding a small value to avoid division by zero
    weights = jnp.exp(-(near_diffs**2) / (2 * sigma**2)) + 1e-16
    weights /= weights.sum()

    # Manually shift each PSF
    shifted_psf1, mask1 = shift_and_mask(
        near_psfs[0],
        near_shifts[0, 0],
        near_shifts[0, 1],
        weights[0],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )
    shifted_psf2, mask2 = shift_and_mask(
        near_psfs[1],
        near_shifts[1, 0],
        near_shifts[1, 1],
        weights[1],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )

    # Accumulate the weighted PSFs and the weight masks
    psf = shifted_psf1 + shifted_psf2
    weight_array = mask1 + mask2

    # Normalize the PSF
    safe_reciprocal = jnp.where(weight_array != 0, 1.0 / weight_array, 0.0)
    psf = psf * safe_reciprocal

    return psf


def create_avg_psf_2DQ(
    x: float,
    y: float,
    pixel_scale: float,
    x_offsets: jnp.ndarray,
    y_offsets: jnp.ndarray,
    x_grid: jnp.ndarray,
    y_grid: jnp.ndarray,
    x_phasor: jnp.ndarray,
    y_phasor: jnp.ndarray,
    reshaped_psfs: jnp.ndarray,
):
    """Creates and returns the PSF at the specified off-axis position using JAX."""
    # The core logic is similar to OffAx, but uses JAX operations
    _x, _y = convert_xy_2DQ(x, y)

    near_inds, near_offsets = get_near_inds_offsets_2D(x_offsets, y_offsets, _x, _y)

    near_psfs = reshaped_psfs[near_inds[:, 0], near_inds[:, 1]]

    near_shifts = (jnp.array([_x, _y]) - near_offsets) / pixel_scale
    near_diffs = jnp.linalg.norm(near_shifts, axis=1)
    sigma = 0.25
    # Adding a small value to avoid division by zero
    weights = jnp.exp(-(near_diffs**2) / (2 * sigma**2)) + 1e-16
    weights /= weights.sum()

    # Manually shift each PSF
    shifted_psf1, mask1 = shift_and_mask(
        near_psfs[0],
        near_shifts[0, 0],
        near_shifts[0, 1],
        weights[0],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )
    shifted_psf2, mask2 = shift_and_mask(
        near_psfs[1],
        near_shifts[1, 0],
        near_shifts[1, 1],
        weights[1],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )
    shifted_psf3, mask3 = shift_and_mask(
        near_psfs[2],
        near_shifts[2, 0],
        near_shifts[2, 1],
        weights[2],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )
    shifted_psf4, mask4 = shift_and_mask(
        near_psfs[3],
        near_shifts[3, 0],
        near_shifts[3, 1],
        weights[3],
        x_grid,
        y_grid,
        x_phasor,
        y_phasor,
    )

    # Accumulate the weighted PSFs and the weight masks
    psf = shifted_psf1 + shifted_psf2 + shifted_psf3 + shifted_psf4
    weight_array = mask1 + mask2 + mask3 + mask4

    safe_reciprocal = jnp.where(weight_array != 0, 1.0 / weight_array, 0.0)
    psf = psf * safe_reciprocal

    return psf


def get_near_inds_offsets_2D(
    x_offsets: jnp.ndarray, y_offsets: jnp.ndarray, _x: float, _y: float
):
    """Computes the nearest indices and offsets for the given (_x, _y) position in 2D.

    Args:
        x_offsets (jnp.ndarray): 1D array of x offsets, sorted in ascending order.
        y_offsets (jnp.ndarray): 1D array of y offsets, sorted in ascending order.
        _x (float): The x-coordinate for which to find nearby offsets.
        _y (float): The y-coordinate for which to find nearby offsets.

    Returns:
        Tuple[jnp.ndarray, jnp.ndarray]:
            - near_inds: Array of shape (4, 2) containing the indices of the
              four surrounding points.
            - near_offsets: Array of shape (4, 2) containing the corresponding
              (x, y) offsets.
    """
    # Find insertion indices for _x and _y
    _x_ind = jnp.searchsorted(x_offsets, _x, side="left")
    _y_ind = jnp.searchsorted(y_offsets, _y, side="left")

    # Handle boundary conditions for x indices
    x_ind_low = jnp.clip(_x_ind - 1, 0, x_offsets.size - 1)
    x_ind_high = jnp.clip(_x_ind, 0, x_offsets.size - 1)

    # Handle boundary conditions for y indices
    y_ind_low = jnp.clip(_y_ind - 1, 0, y_offsets.size - 1)
    y_ind_high = jnp.clip(_y_ind, 0, y_offsets.size - 1)

    # Collect the two nearest indices for x and y
    x_inds = jnp.array([x_ind_low, x_ind_high])
    y_inds = jnp.array([y_ind_low, y_ind_high])

    # Manually create the four combinations without using meshgrid
    near_inds = jnp.array(
        [
            [x_inds[0], y_inds[0]],
            [x_inds[0], y_inds[1]],
            [x_inds[1], y_inds[0]],
            [x_inds[1], y_inds[1]],
        ]
    )

    # Extract the corresponding x and y offset values
    x_vals_low = x_offsets[x_inds[0]]
    x_vals_high = x_offsets[x_inds[1]]
    y_vals_low = y_offsets[y_inds[0]]
    y_vals_high = y_offsets[y_inds[1]]

    # Manually create the corresponding (x, y) offset combinations
    near_offsets = jnp.array(
        [
            [x_vals_low, y_vals_low],
            [x_vals_low, y_vals_high],
            [x_vals_high, y_vals_low],
            [x_vals_high, y_vals_high],
        ]
    )
    return near_inds, near_offsets


def shift_and_mask(
    near_psf, shift_x, shift_y, weight, x_grid, y_grid, x_phasor, y_phasor
):
    """Shifts the PSF in x and y directions and applies a weight mask.

    Args:
        near_psf (jax.numpy.ndarray): The PSF image to shift.
        shift_x (float): Shift in the x-direction.
        shift_y (float): Shift in the y-direction.
        weight (float): Weight for the PSF.
        x_grid (jax.numpy.ndarray): The x-coordinate grid.
        y_grid (jax.numpy.ndarray): The y-coordinate grid.
        x_phasor (jax.numpy.ndarray): Precomputed components for the Fourier shift.
        y_phasor (jax.numpy.ndarray): Precomputed components for the Fourier shift.

    Returns:
        Tuple[jax.numpy.ndarray, jax.numpy.ndarray]:
            - Weighted shifted PSF.
            - Weight mask.
    """
    # Shift the PSF in x and y directions
    shifted_psf = fft_shift_x(near_psf, shift_x, x_phasor)
    shifted_psf = fft_shift_y(shifted_psf, shift_y, y_phasor)

    # Create the weight mask
    weight_mask = create_shift_mask(
        near_psf, shift_x, shift_y, x_grid, y_grid, fill_val=weight
    )

    # Apply the weight mask
    weighted_psf = weight_mask * shifted_psf

    return weighted_psf, weight_mask


def convert_xy_1D(x, y):
    """Converts x and y to 1D coordinates."""
    return jnp.sqrt(x**2 + y**2), 0


def convert_xy_2DQ(x, y):
    """Converts x and y to 2D quarter symmetric coordinates."""
    return jnp.abs(x), jnp.abs(y)


def convert_xy_2D(x, y):
    """Converts x and y to 2D coordinates."""
    return x, y


def basic_shift_val(input_val, converted_val, pixel_scale):
    """Calculates the shift in pixels for a basic shift."""
    return (input_val - converted_val) / pixel_scale


def sym_shift_val(input_val, converted_val, pixel_scale):
    """Calculates the shift in pixels for a symmetric shift."""
    flip = jnp.sign(input_val) == -1
    return lax.cond(
        flip,
        lambda _: (input_val + converted_val) / pixel_scale,
        lambda _: (input_val - converted_val) / pixel_scale,
        operand=None,
    )


def x_basic_shift(input_val, converted_val, PSF, pixel_scale, x_phasor):
    """Shifts the PSF to the specified x position."""
    shift = basic_shift_val(input_val, converted_val, pixel_scale)
    return fft_shift_x(PSF, shift, x_phasor)


def y_basic_shift(input_val, converted_val, PSF, pixel_scale, y_phasor):
    """Shifts the PSF to the specified y position."""
    shift = basic_shift_val(input_val, converted_val, pixel_scale)
    return fft_shift_y(PSF, shift, y_phasor)


def x_symmetric_shift(input_val, converted_val, PSF, pixel_scale, x_phasor):
    """Shifts the PSF to the specified position assuming symmetry about x=0."""
    flip = jnp.sign(input_val) == -1
    # Apply a horizontal flip if the input value is negative
    _PSF = lax.cond(
        flip,
        lambda _: jnp.fliplr(PSF),
        lambda _: PSF,
        operand=None,
    )
    # Calculate the distance to shift the PSF
    shift = sym_shift_val(input_val, converted_val, pixel_scale)
    # Apply the shift
    return fft_shift_x(_PSF, shift, x_phasor)


def y_symmetric_shift(input_val, converted_val, PSF, pixel_scale, y_phasor):
    """Shifts the PSF to the specified position assuming symmetry about y=0."""
    flip = jnp.sign(input_val) == -1
    # Apply a vertical flip if the input value is negative
    _PSF = lax.cond(
        flip,
        lambda _: jnp.flipud(PSF),
        lambda _: PSF,
        operand=None,
    )
    # Get the distance to shift the PSF
    shift = sym_shift_val(input_val, converted_val, pixel_scale)
    # Apply the shift
    return fft_shift_y(_PSF, shift, y_phasor)


def fft_rotate_jax(image, rot_deg):
    """Rotate an image by a specified angle using Fourier-based shear operations.

    This function performs an image rotation by decomposing the rotation into
    three sequential shear operations in the Fourier domain. For more details
    see Larkin et al. (1997).

    Args:
        image (jax.numpy.ndarray):
            The input image to be rotated.
        rot_deg (float):
            The rotation angle in degrees. Positive values rotate the image
            counterclockwise, and negative values rotate it clockwise.

    Returns:
        jax.numpy.ndarray:
            The rotated image.
    """
    # To rotate counterclockwise, with the origin in the lower left, we use the
    # negative of the angle
    rot_deg = -rot_deg

    # Cut the angle to (-45, 45] and a number of 90-degree rotations
    rot_deg, n_rot = decompose_angle_jax(rot_deg)

    image = rot90_helper_jax(image, n_rot)

    image = lax.cond(
        rot_deg != 0.0,
        lambda x: rotate_with_shear(image, x),
        lambda x: image,
        rot_deg,
    )

    return image


def rotate_with_shear(image, rot_deg):
    """Rotate an image by a specified angle using Fourier-based shear operations.

    This is a helper function that simplifies the fft_rotate_jax function by
    simplifying the lambda function used in the lax.cond call.

    Args:
        image (jax.numpy.ndarray):
            The input image to be rotated.
        rot_deg (float):
            The rotation angle in degrees.

    Returns:
        jax.numpy.ndarray:
            The rotated image.
    """
    theta = jnp.deg2rad(rot_deg)
    a = jnp.tan(theta / 2)
    b = -jnp.sin(theta)

    x_freqs, x_dists, y_freqs, y_dists = fft_shear_setup(image)
    # Rotate using three shears
    # s_x
    image = fft_shear_x(image, a, x_freqs, x_dists)

    # s_yx
    image = fft_shear_y(image, b, y_freqs, y_dists)

    # s_xyx
    image = fft_shear_x(image, a, x_freqs, x_dists)
    return image


def fft_shear_setup(image):
    """Perform a shear operation in the Fourier domain.

    Args:
        image (jax.numpy.ndarray):
            The input image to be sheared.

    Returns:
        tuple:
            - jax.numpy.ndarray: x frequencies used for the Fourier transform.
            - jax.numpy.ndarray: x distances from the center of the image.
            - jax.numpy.ndarray: y frequencies used for the Fourier transform.
            - jax.numpy.ndarray: y distances from the center of the image
    """
    # Calculate padding size based on the image dimensions
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)

    # Pad the image with zeros
    padded = jnp.pad(image, n_pad, mode="constant")

    # Calculate the coordinate array for the padded image
    padded_height, padded_width = padded.shape
    center_y, center_x = (jnp.array(padded.shape) - 1) / 2
    grid_y, grid_x = jnp.mgrid[0:padded_height, 0:padded_width]

    # Array of distances from the center of the image along the shear axis
    # if axis == 1:
    # Shearing along the horizontal axis
    x_dists = grid_x - center_x
    x_perpendicular_axis = 1
    # Compute the Fourier frequencies for the dimension perpendicular to the shear axis
    x_freqs = jnp.fft.fftfreq(x_dists.shape[x_perpendicular_axis])
    x_freqs = jnp.fft.fftshift(x_freqs)

    # Tile the shifted frequencies to match the dimensions of the padded image
    x_freqs = jnp.tile(x_freqs, (x_dists.shape[1], 1)).T

    # Shearing along the vertical axis
    y_dists = grid_y - center_y

    # Determine the perpendicular axis to the shear direction
    y_perpendicular_axis = 0

    y_freqs = jnp.fft.fftfreq(y_dists.shape[y_perpendicular_axis])
    y_freqs = jnp.fft.fftshift(y_freqs)
    y_freqs = jnp.tile(y_freqs, (y_dists.shape[0], 1))

    return x_freqs, x_dists, y_freqs, y_dists


def fft_shear_x(image, shear_factor, x_freqs, x_dists):
    """Perform a shear operation in the Fourier domain along the x-axis.

    Uses JAX functions to perform the shear operation in the Fourier domain
    along the x-axis.

    Args:
        image (jax.numpy.ndarray):
            The input image to be sheared.
        shear_factor (float):
            The shear factor.
        x_freqs (jax.numpy.ndarray):
            x frequencies used for the Fourier transform.
        x_dists (jax.numpy.ndarray):
            x distances from the center of the image.

    Returns:
        jax.numpy.ndarray:
            The sheared image with the zero padding removed.
    """
    # Calculate padding size based on the image dimensions
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)
    img_edge = n_pad + n_pixels

    # Pad the image with zeros
    padded = jnp.pad(image, n_pad, mode="constant")
    padded = jnp.fft.fftshift(padded)
    padded = jnp.fft.fftshift(jnp.fft.fft(padded, axis=1))

    # Apply the phase shift (shear) in the Fourier domain
    padded = jnp.exp(-2j * jnp.pi * shear_factor * x_freqs * x_dists) * padded

    # Shift back and apply the inverse Fourier transform along the specified axis
    padded = jnp.fft.fftshift(padded)
    padded = jnp.fft.ifft(padded, axis=1)
    padded = jnp.fft.fftshift(padded)

    # Unpad the image to return to the original size
    image = jnp.real(padded[n_pad:img_edge, n_pad:img_edge])

    return image


def fft_shear_y(image, shear_factor, y_freqs, y_dists):
    """Perform a shear operation in the Fourier domain along the y-axis.

    Uses JAX operations.

    Args:
        image (jax.numpy.ndarray):
            The input image to be sheared.
        shear_factor (float):
            The shear factor.
        y_freqs (jax.numpy.ndarray):
            y frequencies used for the Fourier transform.
        y_dists (jax.numpy.ndarray):
            y distances from the center of the image.

    Returns:
        jax.numpy.ndarray:
            The sheared image with the zero padding removed.
    """
    # Calculate padding size based on the image dimensions
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)
    img_edge = n_pad + n_pixels

    # Pad the image with zeros
    padded = jnp.pad(image, n_pad, mode="constant")
    padded = jnp.fft.fftshift(padded)
    padded = jnp.fft.fftshift(jnp.fft.fft(padded, axis=0))

    # Apply the phase shift (shear) in the Fourier domain
    padded = jnp.exp(-2j * jnp.pi * shear_factor * y_freqs * y_dists) * padded

    # Shift back and apply the inverse Fourier transform along the specified axis
    padded = jnp.fft.fftshift(padded)
    padded = jnp.fft.ifft(padded, axis=0)
    padded = jnp.fft.fftshift(padded)

    # Unpad the image to return to the original size
    image = jnp.real(padded[n_pad:img_edge, n_pad:img_edge])

    return image


def decompose_angle_jax(angle):
    """Decompose an angle from [0, 360) to (-45, 45] and 90 degree rotations.

    Args:
        angle (float):
            The input angle in degrees.

    Returns:
        tuple:
            - float: The rotation angle in the range (-45, 45] degrees.
            - int: The number of 90-degree rotations to apply.
    """
    # Normalize the angle to [0, 360)
    angle = angle % 360

    # Determine the number of 90-degree rotations
    n_rot = (angle // 90).astype(int)

    # Cut the angle to [0, 90)
    adjusted_angle = angle % 90

    # Adjust the angle to the range (-45, 45]
    adjusted_angle, n_rot = lax.cond(
        adjusted_angle > 45,
        lambda x: (x - 90, n_rot + 1),
        lambda x: (x, n_rot),
        adjusted_angle,
    )
    # if n_rot is 4, set it to 0 to avoid unnecessary rotations
    # this occurs when the angle is in the range (315, 360)
    n_rot = lax.cond(
        n_rot == 4,
        lambda x: 0,
        lambda x: x,
        n_rot,
    )

    return adjusted_angle, n_rot


def rot90_traceable(m, k=1, axes=(0, 1)):
    """Rotate an array by 90 degrees in the plane specified by axes.

    This function is a traceable version of `numpy.rot90` taken from the
    jax GitHub issues.

    Args:
        m (jax.numpy.ndarray):
            The input array to be rotated.
        k (int, optional):
            The number of 90-degree rotations to apply.
        axes (tuple, optional):
            The axes to rotate the array in.

    Returns:
        jax.numpy.ndarray:
            The rotated array.
    """
    k %= 4
    return lax.switch(k, [partial(jnp.rot90, m, k=i, axes=axes) for i in range(4)])


def rot90_helper_jax(image, n_rot):
    """Rotate an image by 90 degrees a specified number of times.

    Ensures that the image is of odd dimensions before rotating to avoid
    incorrect centering. Uses JAX operations.

    Args:
        image (jax.numpy.ndarray):
            The input image to be rotated.
        n_rot (int):
            The number of 90-degree rotations to apply.

    Returns:
        jax.numpy.ndarray:
            The rotated image.
    """
    # Check if the image needs padding for even dimensions
    needs_padding = jnp.logical_or(image.shape[0] % 2 == 0, image.shape[1] % 2 == 0)

    def pad_and_rotate(img, n_rot):
        padded = jnp.zeros((img.shape[0] + 1, img.shape[1] + 1))
        padded = padded.at[:-1, :-1].set(img)
        rotated = rot90_traceable(padded, k=n_rot)
        return rotated[:-1, :-1]

    def rotate_only(img, n_rot):
        return rot90_traceable(img, k=n_rot)

    _img = lax.cond(
        needs_padding,
        lambda img: pad_and_rotate(img, n_rot),
        lambda img: rotate_only(img, n_rot),
        image,
    )

    return _img



"""Module for handling off-axis PSFs using JAX."""

from pathlib import Path

import jax.numpy as jnp
from astropy.units import Quantity
from jax import device_put, jit, vmap
from jax.experimental import mesh_utils
from jax.experimental.shard_map import shard_map
from jax.sharding import Mesh
from jax.sharding import PartitionSpec as P

from .jax_funcs import (
    basic_shift_val,
    convert_xy_1D,
    convert_xy_2DQ,
    create_avg_psf_1D,
    create_avg_psf_2DQ,
    create_shift_mask,
    get_pad_info,
    sym_shift_val,
    x_basic_shift,
    x_symmetric_shift,
    y_basic_shift,
    y_symmetric_shift,
)
from .offax import OffAx


class OffJAX(OffAx):
    """Class for handling off-axis PSFs using JAX.

    This class inherits from OffAx and uses JAX for optimized computation.

    Attributes:
        pixel_scale (Quantity):
            Pixel scale of the PSF data in lambda/D.
        center_x (Quantity):
            Central x position in the PSF data.
        center_y (Quantity):
            Central y position in the PSF data.
        reshaped_psfs (jnp.ndarray):
            The PSF data, cast as a JAX array.
    """

    def __init__(
        self,
        yip_dir: Path,
        offax_data_file: str,
        offax_offsets_file: str,
        pixel_scale: Quantity,
        x_symmetric: bool,
        y_symmetric: bool,
        cpu_cores: int = 1,
        platform: str = "cpu",
    ) -> None:
        """Initializes the OffJAX class by casting YIP data to JAX arrays."""
        super().__init__(
            yip_dir,
            offax_data_file,
            offax_offsets_file,
            pixel_scale,
            x_symmetric,
            y_symmetric,
        )
        self.cpu_cores = cpu_cores
        self.platform = platform

        ##############
        # Convert the PSF data to JAX arrays
        ##############
        self.reshaped_psfs = device_put(jnp.array(self.reshaped_psfs))
        n_pixels_orig, n_pad, img_edge, n_pixels_final = get_pad_info(
            self.reshaped_psfs[0, 0], 1.5
        )

        self.x_offsets = device_put(jnp.array(self.x_offsets))
        self.y_offsets = device_put(jnp.array(self.y_offsets))
        # Precompute and store coordinate grids based on PSF shape
        height, width = (self.reshaped_psfs.shape[2], self.reshaped_psfs.shape[3])
        # height, width = (self.padded_psfs.shape[2], self.padded_psfs.shape[3])

        self.x_grid, self.y_grid = jnp.meshgrid(
            jnp.arange(width), jnp.arange(height), indexing="xy"
        )
        self.x_grid = device_put(self.x_grid)
        self.y_grid = device_put(self.y_grid)

        # Create the frequency grids
        ky = jnp.fft.fftfreq(n_pixels_final)
        kx = jnp.fft.fftfreq(n_pixels_final)

        # Precomputed base exponentials
        self.x_phasor = jnp.exp(-2j * jnp.pi * kx)
        self.y_phasor = jnp.exp(-2j * jnp.pi * ky)

        ##############
        # Choose the transformations
        ##############
        if self.type == "1d":
            create_avg_psf = create_avg_psf_1D
            convert_xy = convert_xy_1D
        elif self.type == "2dq":
            create_avg_psf = create_avg_psf_2DQ
            convert_xy = convert_xy_2DQ

        if self.x_symmetric:
            x_shift = x_symmetric_shift
            x_shift_val = sym_shift_val
        else:
            x_shift = x_basic_shift
            x_shift_val = basic_shift_val
        if self.y_symmetric:
            y_shift = y_symmetric_shift
            y_shift_val = sym_shift_val
        else:
            y_shift = y_basic_shift
            y_shift_val = basic_shift_val

        ##############
        # Create the JAX functions
        ##############
        def create_psf(x, y):
            """Create an off-axis PSF at a given position.

            This uses closures to pass in the reshaped PSFs, pixel scale, and
            offsets which allows for JIT compilation that treats them as
            constants.

            Args:
                x (float):
                    x position in lambda/D.
                y (float):
                    y position in lambda/D.

            Returns:
                jnp.ndarray:
                    The off-axis PSF at the given position.
            """
            avg_psf = create_avg_psf(
                x,
                y,
                self.pixel_scale.value,
                self.x_offsets,
                self.y_offsets,
                self.x_grid,
                self.y_grid,
                self.x_phasor,
                self.y_phasor,
                self.reshaped_psfs,
            )
            _x, _y = convert_xy(x, y)
            psf = x_shift(x, _x, avg_psf, self.pixel_scale.value, self.x_phasor)
            psf = y_shift(y, _y, psf, self.pixel_scale.value, self.y_phasor)
            _x_shift = x_shift_val(x, _x, self.pixel_scale.value)
            _y_shift = y_shift_val(y, _y, self.pixel_scale.value)
            mask = create_shift_mask(
                psf, _x_shift, _y_shift, self.x_grid, self.y_grid, 1
            )
            return psf * mask

        self.create_psf = jit(create_psf)
        self.create_psfs = jit(vmap(create_psf, in_axes=(0, 0)))

    def create_psfs_parallel(self, x_vals, y_vals):
        """Create off-axis PSFs at multiple positions in parallel using shard_map.

        If the number of (x, y) pairs doesn't evenly divide the number of devices,
        we pad the inputs and then discard the extra results after processing.

        Args:
            x_vals (jnp.ndarray):
                Array of x-coordinates of shape (N,).
            y_vals (jnp.ndarray):
                Array of y-coordinates of shape (N,).

        Returns:
            jnp.ndarray:
                Array of off-axis PSFs of shape (N, height, width).
        """
        D = self.cpu_cores  # Number of devices
        N = x_vals.shape[0]

        # Determine if we need padding
        remainder = N % D
        padding_needed = (D - remainder) if remainder != 0 else 0

        # Pad inputs so total number of items is divisible by D
        if padding_needed > 0:
            x_vals_padded = jnp.pad(x_vals, (0, padding_needed), constant_values=0)
            y_vals_padded = jnp.pad(y_vals, (0, padding_needed), constant_values=0)
        else:
            x_vals_padded = x_vals
            y_vals_padded = y_vals

        # Set up the device mesh for shard_map
        mesh = Mesh(
            mesh_utils.create_device_mesh([D]),
            axis_names=("i",),
        )

        # Distribute computation across devices. Each device gets N/D items.
        # in_specs=P('i') will shard the first dimension across the 'i' axis.
        psfs_padded = shard_map(
            self.create_psfs,
            mesh=mesh,
            in_specs=P("i"),
            out_specs=P("i"),
        )(x_vals_padded, y_vals_padded)

        # Truncate the padded results if we added any padding
        if padding_needed > 0:
            psfs = psfs_padded[:N]
        else:
            psfs = psfs_padded

        return psfs



"""Utility functions for the yippy package."""

from pathlib import Path

import astropy.io.fits as fits
import astropy.units as u
import numpy as np
from astropy.units import Quantity
from lod_unit import lod, lod_eq
from scipy.ndimage import zoom


def convert_to_lod(
    x: Quantity, center_pix=None, pixel_scale=None, lam=None, D=None, dist=None
) -> Quantity:
    """Convert the x/y position to lambda/D.

    This function has the following assumptions on the x/y values provided:
        - If units are pixels, they follow the 00LL convention. As in the (0,0)
          point is the lower left corner of the image.
        - If the x/y values are in lambda/D, angular, or length units the
            (0,0) point is the center of the image, where the star is
            (hopefully) located.

    Args:
        x (astropy.units.Quantity):
            Position. Can be units of pixel, an angular unit (e.g. arcsec),
            or a length unit (e.g. AU)
        center_pix (astropy.units.Quantity):
            Center of the image in pixels (for the relevant axis)
        pixel_scale (astropy.units.Quantity):
            Pixel scale of in
        lam (astropy.units.Quantity):
            Wavelength of the observation
        D (astropy.units.Quantity):
            Diameter of the telescope
        dist (astropy.units.Quantity):
            Distance to the system
    """
    if x.unit == "pixel":
        assert center_pix is not None, (
            "Center pixel must be provided to convert pixel to lod."
        )
        assert pixel_scale is not None, (
            "Pixel scale must be provided to convert pixel to lod."
        )
        assert pixel_scale.unit == (lod / u.pix), (
            f"Pixel scale must be in units of lod/pix, not {pixel_scale.unit}."
        )

        x = x - center_pix
        x = x * pixel_scale
        # Center the x position
    elif x.unit.physical_type == "angle":
        assert lam is not None, (
            f"Wavelength must be provided to convert {x.unit.physical_type} to lod."
        )
        assert D is not None, (
            f"Telescope diameter must be provided to convert {x.unit.physical_type}"
            f" to lod."
        )
        x = x.to(lod, lod_eq(lam, D))
    elif x.unit.physical_type == "length":
        # If the distance to the system is not provided, raise an error
        assert dist is not None, (
            f"Distance to system must be provided to convert {x.unit.physical_type}"
            f" to {lod}."
        )
        x_angular = np.arctan(x.to(u.m).value / dist.to(u.m).value) * u.rad
        x = x_angular.to(lod, lod_eq(lam, D))
    else:
        raise ValueError(f"No conversion implemented for {x.unit.physical_type}")
    return x


def convert_to_pix(
    x: Quantity, center_pix, pixel_scale, lam=None, D=None, dist=None
) -> Quantity:
    """Convert the x/y position from lambda/D to pixel units.

    This function has the following assumptions on the x/y values provided:
        - If the desired output is in pixels, the (0,0) point is the lower left
          corner of the image.
        - If the x/y values are in lambda/D, angular, or length units, the
          (0,0) point is the center of the image, where the star is
          (hopefully) located.

    Args:
        x (astropy.units.Quantity):
            Position to convert. Should be in units of lambda/D, an angular unit
            (e.g., arcsec), or a length unit (e.g., AU).
        center_pix (astropy.units.Quantity, optional):
            Center of the image in pixels (for the relevant axis). Required if
            converting to pixel units.
        pixel_scale (astropy.units.Quantity, optional):
            Pixel scale in units of lambda/D per pixel. Required if converting to
            pixel units.
        lam (astropy.units.Quantity, optional):
            Wavelength of the observation. Required if converting from angular or
            length units to pixel units.
        D (astropy.units.Quantity, optional):
            Diameter of the telescope. Required if converting from angular or
            length units to pixel units.
        dist (astropy.units.Quantity, optional):
            Distance to the system. Required if converting from length units to
            pixel units.

    Returns:
        astropy.units.Quantity:
            Position in pixel units.

    Raises:
        AssertionError:
            If required parameters for the conversion are not provided or have
            incorrect units.
        ValueError:
            If the input unit type is not supported for conversion.
    """
    if isinstance(x, float) or isinstance(x, np.floating):
        # Assume x is a float in lambda/D
        x_pixels = x * lod / pixel_scale + center_pix
    elif x.unit == lod:
        # Center the x position
        x_pixels = x / pixel_scale + center_pix

    elif x.unit.physical_type == "angle":
        # Conversion from angle to pixels
        assert lam is not None, (
            "Wavelength must be provided to convert angle to pixels."
        )
        assert D is not None, (
            "Telescope diameter must be provided to convert angle to pixels."
        )

        # Convert angle to lambda/D
        x_lod = x.to(u.rad, lod_eq(lam, D))
        # Now convert lambda/D to pixels
        x_pixels = x_lod / pixel_scale + center_pix

    elif x.unit.physical_type == "length":
        # Conversion from length to pixels
        assert lam is not None, (
            "Wavelength must be provided to convert length to pixels."
        )
        assert D is not None, (
            "Telescope diameter must be provided to convert length to pixels."
        )
        assert dist is not None, (
            "Distance to system must be provided to convert length to pixels."
        )

        # Convert length to angle
        x_angle = np.arctan(x.to(u.m).value / dist.to(u.m).value) * u.rad
        # Convert angle to lambda/D
        x_lod = x_angle.to(lod, lod_eq(lam, D))
        # Now convert lambda/D to pixels
        x_pixels = x_lod / pixel_scale + center_pix

    else:
        raise ValueError(f"No conversion implemented for {x.unit.physical_type}")

    return x_pixels


def save_coro_performance_to_fits(
    sep: np.ndarray,
    throughput: np.ndarray,
    raw_contrast: np.ndarray,
    filename: str,
    outdir: Path,
    overwrite=True,
):
    """Save coronagraph performance (throughput, raw_contrast) to a FITS file."""
    sort_idx = np.argsort(sep)
    col_sep = fits.Column(name="separation_lamD", format="E", array=sep[sort_idx])
    col_thr = fits.Column(name="throughput", format="E", array=throughput[sort_idx])
    col_con = fits.Column(name="raw_contrast", format="E", array=raw_contrast[sort_idx])
    cols = fits.ColDefs([col_sep, col_thr, col_con])
    tbhdu = fits.BinTableHDU.from_columns(cols)
    tbhdu.name = "CORO_PERFORMANCE"

    primary_hdu = fits.PrimaryHDU()
    hdul = fits.HDUList([primary_hdu, tbhdu])
    outpath = outdir / filename
    hdul.writeto(outpath, overwrite=overwrite)


def load_coro_performance_from_fits(filename: str, indir: Path):
    """Load (separation, throughput, raw_contrast) from a FITS file."""
    inpath = indir / filename
    with fits.open(inpath) as hdul:
        data = hdul["CORO_PERFORMANCE"].data
        sep = data["separation_lamD"]
        thr = data["throughput"]
        con = data["raw_contrast"]
    return sep, thr, con


def extract_and_oversample_subarray(
    psf_img: np.ndarray,
    center_x: float,
    center_y: float,
    radius_pix: float,
    oversample: int,
):
    """Get oversampled subarray of the PSF image around a given center.

    Extract a subarray of `psf_img` around (center_x, center_y),
    then oversample that subarray by the specified factor.

    Args:
        psf_img (np.ndarray):
            The input PSF image
        center_x (float):
            Position of the center in the x direction
        center_y (float):
            Position of the center in the y direction
        radius_pix (float):
            The radius of the subarray in pixels
        oversample (int):
            The oversampling factor

    Returns:
        subarr_oversamp (np.ndarray):
            The oversampled subarray
        center_x_os (float):
            center_x in oversampled subarray coords
        center_y_os (float):
            center_y in oversampled subarray coords
        radius_os (float):
            radius_pix * oversample
        subarr (np.ndarray):
            the original subarray (for flux renormalization)
    """
    ny, nx = psf_img.shape
    margin = int(np.ceil(radius_pix * 3))
    xmin = max(0, int(np.floor(center_x - margin)))
    xmax = min(nx - 1, int(np.ceil(center_x + margin)))
    ymin = max(0, int(np.floor(center_y - margin)))
    ymax = min(ny - 1, int(np.ceil(center_y + margin)))

    subarr = psf_img[ymin : ymax + 1, xmin : xmax + 1]

    subarr_oversamp = zoom(subarr, oversample, order=1)

    center_x_os = (center_x - xmin) * oversample
    center_y_os = (center_y - ymin) * oversample
    radius_os = radius_pix * oversample

    return subarr_oversamp, center_x_os, center_y_os, radius_os, subarr


def measure_flux_in_oversampled_aperture(
    subarr_oversamp: np.ndarray,
    center_x_os: float,
    center_y_os: float,
    radius_os: float,
    subarr_original: np.ndarray,
) -> float:
    """Get flux in a circular aperture of radius `radius_os` in the oversampled array.

    Returns:
        flux_in_ap (float): total flux inside the circular mask
    """
    yy_os, xx_os = np.indices(subarr_oversamp.shape)
    rr_os = np.sqrt((xx_os - center_x_os) ** 2 + (yy_os - center_y_os) ** 2)
    ap_mask = rr_os <= radius_os

    flux_in_ap = subarr_oversamp[ap_mask].sum()

    # Renormalize flux because zoom can change total flux slightly.
    total_orig = subarr_original.sum()
    total_os = subarr_oversamp.sum()
    if total_os > 0:
        flux_in_ap *= total_orig / total_os

    return flux_in_ap


def fft_shift(image, x=0, y=0):
    """Apply a Fourier shift to an image along the x and/or y axes.

    This function performs a 1D Fourier shift along the x-axis and/or y-axis
    of the input image. If a shift of 0 is specified for either axis, no
    operation is performed along that axis.

    The function uses the `fft_shift_1d` method to apply the shift in the
    Fourier domain, allowing for subpixel accuracy in the shift. The image
    is padded with zeros during the process, and the padding is removed before
    returning the shifted image.

    Args:
        image (numpy.ndarray):
            The input 2D image to be shifted.
        x (float):
            The number of pixels by which to shift the image along the x-axis.
        y (float):
            The number of pixels by which to shift the image along the y-axis.

    Returns:
        numpy.ndarray:
            The shifted image after applying the Fourier transform-based shift
            along the specified axes.
    """
    assert x != 0 or y != 0, "One of x or y must be non-zero."

    if x != 0:
        # Horizontal shift
        image = fft_shift_1d(image, x, axis=1)
    if y != 0:
        # Vertical shift
        image = fft_shift_1d(image, y, axis=0)

    return image


def fft_shift_1d(image, shift_pixels, axis):
    """Apply a Fourier shift to an image along a specified axis.

    This function pads the input image with zeros, performs a 1D Fourier
    transform along the specified axis, shifts the image by a specified
    number of pixels using a phasor, and reconstructs the image after
    applying the shift. The padded regions are removed before returning
    the shifted image.

    Args:
        image (numpy.ndarray):
            The input 2D image to be shifted.
        shift_pixels (float):
            The number of pixels by which to shift the image along the specified axis.
        axis (int):
            The axis to shift (0 for vertical, 1 for horizontal).

    Returns:
        numpy.ndarray:
            The shifted image after applying the Fourier transform and
            removing the padding.
    """
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)
    img_edge = n_pad + n_pixels

    # Pad the image with zeros
    padded = np.pad(image, n_pad, mode="constant")

    # Take the 1D Fourier transform along the specified axis
    padded = np.fft.fft(padded, axis=axis)

    # Get the frequencies used for the Fourier transform along the specified axis
    freqs = np.fft.fftfreq(4 * n_pixels)

    # Create the phasor
    phasor = np.exp(-2j * np.pi * freqs * shift_pixels)

    # Tile the phasor to match the dimensions of the padded image
    if axis == 1:
        phasor = np.tile(phasor, (padded.shape[0], 1))  # Horizontal shift (x-axis)
    else:
        phasor = np.tile(phasor, (padded.shape[1], 1)).T  # Vertical shift (y-axis)

    # Apply the phasor along the specified axis
    padded = padded * phasor

    # Reconstruct the image using the inverse Fourier transform along the specified axis
    padded = np.real(np.fft.ifft(padded, axis=axis))

    # Unpad the image to return to the original size
    image = padded[n_pad:img_edge, n_pad:img_edge]

    return image


def fft_shift_2d(image, x=0, y=0):
    """Apply a 2D Fourier shift to an image along both the x and y axes.

    This function performs a 2D Fourier shift, allowing for subpixel accuracy
    in both x and y directions simultaneously. The image is transformed to the
    frequency domain, shifted by applying a phasor, and then transformed back
    to the spatial domain.

    Args:
        image (numpy.ndarray):
            The input 2D image to be shifted.
        x (float):
            The number of pixels by which to shift the image along the x-axis.
        y (float):
            The number of pixels by which to shift the image along the y-axis.

    Returns:
        numpy.ndarray:
            The shifted image after applying the 2D Fourier transform-based shift.
    """
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)
    img_edge = n_pad + n_pixels

    # Pad the image with zeros
    image = np.pad(image, n_pad, mode="constant")

    # Compute the 2D Fourier transform of the image
    image_ft = np.fft.fft2(image)

    # Get the shape of the image
    n_rows, n_cols = image.shape

    # Create frequency grids for both axes
    ky = np.fft.fftfreq(n_rows)
    kx = np.fft.fftfreq(n_cols)

    # Create meshgrids for frequencies
    # Kx, Ky = np.meshgrid(kx, ky)

    # Create the combined phasor for both shifts
    # phasor = np.exp(-2j * np.pi * (Kx * x + Ky * y))
    # Create 1D phasors
    exp_kx = np.exp(-2j * np.pi * kx * x)
    exp_ky = np.exp(-2j * np.pi * ky * y)
    # Compute outer product for phasor without full meshgrid
    phasor = np.outer(exp_ky, exp_kx)

    # Apply the phasor to the Fourier transformed image
    image_ft_shifted = image_ft * phasor

    # Compute the inverse 2D Fourier transform
    shifted_image = np.fft.ifft2(image_ft_shifted)

    # Return the real part of the shifted image
    shifted_image = np.real(shifted_image)

    return shifted_image[n_pad:img_edge, n_pad:img_edge]


def fft_rotate(image, rot_deg):
    """Rotate an image by a specified angle using Fourier-based shear operations.

    This function performs an image rotation by decomposing the rotation into
    three sequential shear operations in the Fourier domain. For more details
    see Larkin et al. (1997).

    Args:
        image (numpy.ndarray):
            The input image to be rotated.
        rot_deg (float):
            The rotation angle in degrees. Positive values rotate the image
            counterclockwise, and negative values rotate it clockwise.

    Returns:
        numpy.ndarray:
            The rotated image.
    """
    # To rotate counterclockwise, with the origin in the lower left, we use the
    # negative of the angle
    rot_deg = -rot_deg

    # Cut the angle to (-45, 45] and a number of 90-degree rotations
    rot_deg, n_rot = decompose_angle(rot_deg)

    image = rot90_helper(image, n_rot)

    if rot_deg != 0.0:
        theta = np.deg2rad(rot_deg)
        a = np.tan(theta / 2)
        b = -np.sin(theta)

        # Rotate using three shears
        # s_x
        image = fft_shear(image, a, axis=1)

        # s_yx
        image = fft_shear(image, b, axis=0)

        # s_xyx
        image = fft_shear(image, a, axis=1)

    return image


def fft_shear(image, shear_factor, axis):
    """Perform a shear operation in the Fourier domain.

    Args:
        image (numpy.ndarray):
            The input image to be sheared.
        shear_factor (float):
            The shear factor.
        axis (int):
            The axis to shear (0 for vertical, 1 for horizontal).

    Returns:
        numpy.ndarray:
            The sheared image with the zero padding removed.
    """
    # Calculate padding size based on the image dimensions
    n_pixels = image.shape[0]
    n_pad = int(1.5 * n_pixels)
    img_edge = n_pad + n_pixels

    # Pad the image with zeros
    padded = np.pad(image, n_pad, mode="constant")

    # Calculate the coordinate array for the padded image
    padded_height, padded_width = padded.shape
    center_y, center_x = (np.array(padded.shape) - 1) / 2
    grid_y, grid_x = np.mgrid[0:padded_height, 0:padded_width]

    # Array of distances from the center of the image along the shear axis
    if axis == 1:
        # Shearing along the horizontal axis
        dists = grid_x - center_x
    else:
        # Shearing along the vertical axis
        dists = grid_y - center_y

    # Determine the perpendicular axis to the shear direction
    perpendicular_axis = 1 - axis % 2

    # Compute the Fourier frequencies for the dimension perpendicular to the shear axis
    freqs = np.fft.fftfreq(dists.shape[perpendicular_axis])
    freqs = np.fft.fftshift(freqs)

    # Tile the shifted frequencies to match the dimensions of the padded image
    freqs = np.tile(freqs, (dists.shape[axis], 1))

    # Transpose the frequency array if the shear is applied along the horizontal axis
    if axis == 1:
        freqs = freqs.T

    # Shift the padded image to center the zero-frequency component
    padded = np.fft.fftshift(padded)

    # Apply the Fourier transform along the specified axis
    padded = np.fft.fft(padded, axis=axis)
    padded = np.fft.fftshift(padded)

    # Apply the phase shift (shear) in the Fourier domain
    padded = np.exp(-2j * np.pi * shear_factor * freqs * dists) * padded

    # Shift back and apply the inverse Fourier transform along the specified axis
    padded = np.fft.fftshift(padded)
    padded = np.fft.ifft(padded, axis=axis)
    padded = np.fft.fftshift(padded)

    # Unpad the image to return to the original size
    image = np.real(padded[n_pad:img_edge, n_pad:img_edge])

    return image


def rot90_helper(image, n_rot):
    """Rotate an image by 90 degrees a specified number of times.

    Ensures that the image is of odd dimensions before rotating to avoid
    incorrect centering.

    Args:
        image (numpy.ndarray):
            The input image to be rotated.
        n_rot (int):
            The number of 90-degree rotations to apply.

    Returns:
        numpy.ndarray:
            The rotated image.
    """
    # Add padding if the image has even dimensions before rotating
    # by 90 degrees to avoid incorrect centering
    needs_padding = not image.shape[0] % 2 or not image.shape[1] % 2
    if needs_padding:
        _img = np.zeros([image.shape[0] + 1, image.shape[1] + 1])
        _img[:-1, :-1] = image.copy()
    else:
        _img = image.copy()

    # Apply the rotation
    _img = np.rot90(_img, k=n_rot)

    # Remove padding if it was added
    if needs_padding:
        _img = _img[:-1, :-1]

    return _img


def decompose_angle(angle):
    """Decompose an angle from [0, 360) to (-45, 45] and 90 degree rotations.

    Args:
        angle (float):
            The input angle in degrees.

    Returns:
        tuple:
            - float: The rotation angle in the range (-45, 45] degrees.
            - int: The number of 90-degree rotations to apply.
    """
    # Normalize the angle to [0, 360)
    angle = angle % 360

    # Determine the number of 90-degree rotations
    n_rot = int(angle // 90)

    # Cut the angle to [0, 90)
    adjusted_angle = angle % 90

    # Adjust the angle to the range (-45, 45]
    if adjusted_angle > 45:
        adjusted_angle -= 90
        n_rot += 1

    # if n_rot is 4, set it to 0 to avoid unnecessary rotations
    # this occurs when the angle is in the range (315, 360)
    if n_rot == 4:
        n_rot = 0

    return adjusted_angle, n_rot


def create_shift_mask(psf, shift_x, shift_y, fill_val=1):
    """Create a mask to identify valid pixels to average.

    This function is useful because when the PSF is shifted there are empty
    pixels, since they were outside the initial image, and should not be
    included in the final average.

    Args:
        psf (np.ndarray):
            The PSF image to shift.
        shift_x (float):
            The shift in the x direction.
        shift_y (float):
            The shift in the y direction.
        fill_val (float, optional):
            The value to fill the mask with.

    Returns:
        np.ndarray:
            The mask to identify valid pixels to average.
    """
    mask = np.full_like(psf, fill_val)

    # Handle x-direction shifting
    if shift_x > 0:
        # Zero out the left side
        mask[:, : int(np.ceil(shift_x))] = 0
    elif shift_x < 0:
        # Zero out the right side
        mask[:, int(np.floor(shift_x)) :] = 0

    # Handle y-direction shifting
    if shift_y > 0:
        # Zero out the bottom side
        mask[: int(np.ceil(shift_y)), :] = 0
    elif shift_y < 0:
        # Zero out the top side
        mask[int(np.floor(shift_y)) :, :] = 0

    return mask


def fft_zoom(image, zoom_factor):
    """Zoom/rescale an image using Fourier interpolation.

    Args:
        image (np.ndarray):
            The input 2D image to be zoomed.
        zoom_factor (float):
            The factor by which to zoom the image.

    Returns:
        np.ndarray: The resampled image.
    """
    # Perform 2D Fourier transform on the input image
    f_transform = np.fft.fftshift(np.fft.fft2(np.fft.ifftshift(image)))

    # Get the current shape of the image
    old_shape = np.array(image.shape)

    # Calculate the new shape based on the zoom factor
    new_shape = (np.array(image.shape) * zoom_factor).astype(int)

    # Initialize the output Fourier transform with zeros
    f_transform_new = np.zeros(new_shape, dtype=complex)

    # Determine the size of the region to copy from the old Fourier transform
    min_shape = np.minimum(old_shape, new_shape)

    # Define the start and end indices for copying the Fourier transform data
    start_old = (old_shape - min_shape) // 2
    end_old = start_old + min_shape

    start_new = (new_shape - min_shape) // 2
    end_new = start_new + min_shape

    # Copy the relevant portion of the old Fourier transform into the new one
    f_transform_new[start_new[0] : end_new[0], start_new[1] : end_new[1]] = f_transform[
        start_old[0] : end_old[0], start_old[1] : end_old[1]
    ]

    # Perform the inverse Fourier transform to get the zoomed image
    zoomed_image = np.fft.fftshift(np.fft.ifft2(np.fft.ifftshift(f_transform_new)))

    # Return the real part of the resampled image
    return np.real(zoomed_image)


def centered_fft_resample(image, image_pix_scale, new_shape, new_pix_scale):
    """Resample an image from one pixel scale to another using Fourier interpolation.

    This function adjusts the image to a new pixel scale and shape based on the
    given scales and desired output shape, while keeping the image centered.

    Args:
        image (np.ndarray):
            The single image to be resampled.
        image_pix_scale (float):
            The pixel scale of the input image (e.g., arcseconds per pixel).
        new_shape (tuple):
            The desired shape of the output image (height, width).
        new_pix_scale (float):
            The pixel scale of the output image (e.g., arcseconds per pixel).

    Returns:
        np.ndarray: The resampled single image.
    """
    # Calculate the zoom factor for resampling
    zoom_factor = image_pix_scale / new_pix_scale

    # Resample the image using Fourier interpolation
    scaled_image = fft_zoom(image, zoom_factor)

    # Calculate how much the resampled image needs to be cropped or padded
    # to match the desired detector shape
    scaled_shape = np.array(scaled_image.shape)
    new_shape = np.array(new_shape)
    center_offset = (scaled_shape - new_shape) // 2

    # Ensure symmetric padding/cropping
    pad_amount = np.abs(center_offset).astype(int)
    final_image = np.zeros(new_shape, dtype=scaled_image.dtype)

    # Apply padding or cropping based on center_offset
    if np.any(center_offset < 0):
        # Pad the image symmetrically
        padded_image = np.pad(
            scaled_image,
            ((pad_amount[0], pad_amount[0]), (pad_amount[1], pad_amount[1])),
            mode="constant",
        )
        final_image = padded_image[: new_shape[0], : new_shape[1]]
    else:
        # Crop the image symmetrically
        final_image = scaled_image[
            center_offset[0] : center_offset[0] + new_shape[0],
            center_offset[1] : center_offset[1] + new_shape[1],
        ]

    return final_image



"""Module for handling FITS header data."""

import re
from dataclasses import dataclass
from typing import Optional

import astropy.units as u
from astropy.io import fits
from lod_unit import lod

from .logger import logger


@dataclass(frozen=True)
class HeaderData:
    """A dataclass for storing the header data with units attached.

    I think this will be a useful system, however this has not been tested
    considerably. This is just a suggestion for a more robust way to keep the
    FITS header data attached to the coronagraph. Right now, the header data is
    loaded by Coronagraph from the stellar_intens file. This may make more
    sense if Offax, StellarIntens, and SkyTrans each load their own header.
    """

    simple: bool
    bitpix: int
    naxis: int
    naxis1: int
    naxis2: int
    naxis3: int
    design: str
    diameter: Optional[u.Quantity] = None
    pixscale: Optional[u.Quantity] = None
    lambda0: Optional[u.Quantity] = None
    minlam: Optional[u.Quantity] = None
    maxlam: Optional[u.Quantity] = None
    xcenter: Optional[float] = None
    ycenter: Optional[float] = None
    obscured: Optional[float] = None
    jitter: Optional[u.Quantity] = None
    n_lam: Optional[int] = None
    n_star: Optional[int] = None
    zernike: Optional[str] = None
    wfe: Optional[u.Quantity] = None

    @staticmethod
    def extract_unit(comment: str, default_unit: u.Unit, key: str) -> u.Unit:
        """Extract a unit from the comment string of a FITS header entry.

        This attempts to read the FITS header comment and search for an astropy
        Unit. This seems fairly robust, but it may not catch all cases in which
        case a warning is raised. It is also possible that the unit is not
        explicitly stated in the comment, in which case a default unit is
        returned.

        Args:
            comment (str):
                The comment string associated with a FITS header key.
            default_unit (u.Unit):
                The default unit to return if no unit is found in the comment.
            key (str):
                The key of the header entry to extract the unit from.

        Returns:
            u.Unit:
                The extracted astropy unit or the default unit if no
                recognizable unit is found.

        Raises:
            Warning:
                Logs a warning if no unit could be extracted and the default
                unit is used.
        """
        # Define a regex pattern for commonly expected units
        unit_patterns = {
            "nm": u.nm,
            "micron": u.micron,
            "microns": u.micron,
            "um": u.um,
            "mm": u.mm,
            "cm": u.cm,
            "meter": u.m,
            "m": u.m,
            "arcsec": u.arcsec,
            "arcmin": u.arcmin,
            "deg": u.deg,
            "mas": u.mas,
            "pm": u.pm,
        }

        # Match for patterns that look like compound units, e.g., "mas/pixel"
        compound_unit_pattern = r"\b(\w+)/(\w+)\b"
        match = re.search(compound_unit_pattern, comment, re.IGNORECASE)
        if match:
            num_unit, den_unit = match.groups()
            num_astropy_unit = unit_patterns.get(num_unit.lower(), None)
            den_astropy_unit = unit_patterns.get(den_unit.lower(), None)
            if num_astropy_unit and den_astropy_unit:
                logger.debug(
                    f"Extracted compound unit {num_astropy_unit}/{den_astropy_unit}"
                    f' for {key} from "{comment}"'
                )
                return num_astropy_unit / den_astropy_unit

        # Search for any of these single units in the comment
        for pattern, unit in unit_patterns.items():
            if re.search(r"\b" + re.escape(pattern) + r"\b", comment, re.IGNORECASE):
                logger.debug(f'Extracted {unit} from "{comment}" for {key}')
                return unit

        logger.warning(
            (
                f"Using default unit for {key}: {default_unit}. "
                f'Could not extract unit from comment: "{comment}"'
            )
        )
        return default_unit

    @staticmethod
    def get_header_value(header: fits.Header, key: str, default_unit: u.Unit):
        """Retrieves a header value by key and converts it to a specified unit.

        Args:
            header (fits.Header):
                The FITS header from which to retrieve the value.
            key (str):
                The key of the header entry to retrieve.
            default_unit (u.Unit):
                The unit to associate with the returned value.

        Returns:
            Optional[u.Quantity]:
                The header value with the specified unit, or None if the key is
                not present.
        """
        if key in header:
            value = float(header[key])
            unit = HeaderData.extract_unit(header.comments[key], default_unit, key)
            return value * unit
        else:
            return None

    @staticmethod
    def from_fits_header(header: fits.Header):
        """Parses a FITS header.

        This parses the FITS header to initialize the HeaderData class,
        checking for any unhandled fields.

        Args:
            header (fits.Header):
                The FITS header to parse.

        Returns:
            HeaderData:
                An initialized HeaderData object populated with values from the
                FITS header.

        Raises:
            Warning:
                Logs a warning for any unhandled fields in the FITS header.
        """
        recognized_keys = set(
            [
                "SIMPLE",
                "BITPIX",
                "NAXIS",
                "NAXIS1",
                "NAXIS2",
                "NAXIS3",
                "DESIGN",
                "D",
                "PIXSCALE",
                "LAMBDA",
                "MINLAM",
                "MAXLAM",
                "XCENTER",
                "YCENTER",
                "OBSCURED",
                "JITTER",
                "N_LAM",
                "N_STAR",
                "ZERNIKE",
                "WFE",
            ]
        )
        header_keys = set(header.keys())

        unhandled_keys = header_keys - recognized_keys
        if unhandled_keys:
            logger.warning(f"Unhandled header fields: {unhandled_keys}")

        return HeaderData(
            simple=header.get("SIMPLE", "F") == "T",
            bitpix=int(header.get("BITPIX", 0)),
            naxis=int(header.get("NAXIS", 0)),
            naxis1=int(header.get("NAXIS1", 0)),
            naxis2=int(header.get("NAXIS2", 0)),
            naxis3=int(header.get("NAXIS3", 0)),
            design=header.get("DESIGN", ""),
            diameter=HeaderData.get_header_value(header, "D", u.meter),
            pixscale=header.get("PIXSCALE") * lod / u.pix,
            lambda0=HeaderData.get_header_value(header, "LAMBDA", u.micron),
            minlam=HeaderData.get_header_value(header, "MINLAM", u.micron),
            maxlam=HeaderData.get_header_value(header, "MAXLAM", u.micron),
            xcenter=header.get("XCENTER", None),
            ycenter=header.get("YCENTER", None),
            obscured=header.get("OBSCURED", None),
            jitter=HeaderData.get_header_value(header, "JITTER", u.mas),
            n_lam=header.get("N_LAM", None),
            n_star=header.get("N_STAR", None),
            zernike=header.get("ZERNIKE", ""),
            wfe=HeaderData.get_header_value(header, "WFE", u.pm),
        )



"""This module handles the sky transmission map."""

from pathlib import Path

import astropy.io.fits as pyfits


class SkyTrans:
    """Simple class that holds the sky transmission map data.

    This is essentially a placeholder class until I have a better idea of how
    the sky transmission map will be used in the pipeline.

    Attributes:
        data:
            Sky transmission map data.
    """

    def __init__(self, yip_dir: Path, sky_trans_file: str) -> None:
        """Initializes the SkyTrans class by loading the sky transmission map.

        Args:
            yip_dir (Path):
                Path to the directory containing the sky transmission map.
            sky_trans_file (str):
                Name of the file containing the sky transmission map.
        """
        # Load the sky transmission file
        self.data = pyfits.getdata(Path(yip_dir, sky_trans_file), 0)

    def __call__(self):
        """Returns the sky transmission map data."""
        return self.data



"""Base coronagraph class."""

from pathlib import Path

import astropy.io.fits as pyfits
import astropy.units as u
import jax.numpy as jnp
import numpy as np
from lod_unit import lod
from scipy.interpolate import make_interp_spline
from scipy.optimize import root_scalar
from tqdm import tqdm

from .header import HeaderData
from .jax_funcs import enable_x64, set_host_device_count, set_platform
from .logger import logger
from .offax import OffAx
from .offjax import OffJAX
from .sky_trans import SkyTrans
from .stellar_intens import StellarIntens
from .util import (
    convert_to_pix,
    extract_and_oversample_subarray,
    load_coro_performance_from_fits,
    measure_flux_in_oversampled_aperture,
    save_coro_performance_to_fits,
)


class Coronagraph:
    """Primary object for simulating a coronagraph.

    The Coronagraph object manages the coronagraph response for both on-axis
    and off-axis sources. It is primarily called with either
    `:ref:pydata:Coronagraph.offax(x,y)`, to get the off-axis response at a
    given (x,y) offset from the star, or `Coronagraph.stellar(r)` to get the
    coronagraph response at a given stellar angular diameter r.
    """

    def __init__(
        self,
        yip_path: Path,
        use_jax: bool = True,
        use_x64: bool = False,
        stellar_intens_file: str = "stellar_intens.fits",
        stellar_diam_file: str = "stellar_intens_diam_list.fits",
        offax_data_file: str = "offax_psf.fits",
        offax_offsets_file: str = "offax_psf_offset_list.fits",
        sky_trans_file: str = "sky_trans.fits",
        performance_file: str = "coro_perf.fits",
        x_symmetric: bool = True,
        y_symmetric: bool = False,
        cpu_cores: int = 4,
        platform: str = "cpu",
    ):
        """Initialize the Coronagraph object.

        Loads the coronagraph data from the given yield input package and creates
        the interpolation functions for the stellar intensity and off-axis PSF.

        Args:
            yip_path (Path):
                Yield input package directory. Must have fits files
                    offax_psf_offset_list - The off-axis PSF list
                    offax_psf - PSF of off-axis sources
                    sky_trans - Sky transmission data
            use_jax (bool):
                Whether to use JAX for optimized computation. Default is False.
            use_x64 (bool):
                Whether to use 64-bit floating point precision. Default is False.
            stellar_intens_file (str):
                Name of the stellar intensity file. Default is stellar_intens.fits
            stellar_diam_file (str):
                Name of the stellar intensity diameter list file. Default is
                stellar_intens_diam_list.fits
            offax_data_file (str):
                Name of the off-axis PSF file. Default is offax_psf.fits
            offax_offsets_file (str):
                Name of the off-axis PSF offset list file. Default is
                offax_psf_offset_list.fits
            sky_trans_file (str):
                Name of the sky transmission data file. Default is sky_trans.fits.
            performance_file (str):
                Name of the coronagraph performance (contrast, throughput) fits file.
            x_symmetric (bool):
                Whether off-axis PSFs are symmetric about the x-axis. Default is True.
            y_symmetric (bool):
                Whether off-axis PSFs are symmetric about the y-axis. Default is False.
            cpu_cores (int):
                Number of CPU cores to use. Default is 1.
            platform (str):
                Platform to use for JAX computation. Default is "cpu". Options are
                "cpu", "gpu", "tpu".
        """
        ###################
        # Read input data #
        ###################
        yip_path = Path(yip_path)
        self.yip_path = yip_path

        logger.info(f"Creating {yip_path.stem} coronagraph")

        self.name = yip_path.stem
        # Get header and calculate the lambda/D value
        stellar_intens_header = pyfits.getheader(Path(yip_path, stellar_intens_file), 0)

        # Get pixel scale with units
        self.header = HeaderData.from_fits_header(stellar_intens_header)
        self.pixel_scale = self.header.pixscale
        self.frac_obscured = self.header.obscured

        # Stellar intensity of the star being observed as function of stellar
        # angular diameter (unitless)
        self.stellar_intens = StellarIntens(
            yip_path, stellar_intens_file, stellar_diam_file
        )

        # Offaxis PSF of the planet as function of separation from the star
        if use_jax:
            # Apply JAX settings
            if use_x64:
                enable_x64()
            if platform != "cpu":
                set_platform(platform)
            elif cpu_cores > 1:
                set_host_device_count(cpu_cores)
            self.offax = OffJAX(
                yip_path,
                offax_data_file,
                offax_offsets_file,
                self.pixel_scale,
                x_symmetric,
                y_symmetric,
                cpu_cores,
                platform,
            )
        else:
            self.offax = OffAx(
                yip_path,
                offax_data_file,
                offax_offsets_file,
                self.pixel_scale,
                x_symmetric,
                y_symmetric,
            )

        # Get the sky_trans mask
        self.sky_trans = SkyTrans(yip_path, sky_trans_file)

        # PSF datacube here is a 4D array of PSFs at each pixel (x psf offset,
        # y psf offset, x, y). Given the computational cost of generating this
        # datacube, it is only generated when needed.
        self.has_psf_datacube = False

        # Shape of the images in the PSFs
        self.psf_shape = np.array([self.header.naxis1, self.header.naxis2])
        assert self.psf_shape[0] == self.psf_shape[1], "PSF must be square"
        self.npixels = self.psf_shape[0]

        # Get the contrast and throughput
        if self.offax.type == "1d":
            perf_path = Path(self.yip_path, performance_file)
            if perf_path.exists():
                sep, throughput, raw_contrast = load_coro_performance_from_fits(
                    performance_file, self.yip_path
                )
            else:
                logger.info("No precomputed performance file found. Computing now...")
                sep_throughput, throughput = self.get_throughput_curve(plot=False)
                sep_contrast, raw_contrast = self.get_contrast_curve(plot=False)

                assert np.all(sep_throughput == sep_contrast), (
                    "Mismatch in separations for performance parameters"
                )
                sep = sep_throughput

                # Save to fits
                save_coro_performance_to_fits(
                    sep, throughput, raw_contrast, performance_file, self.yip_path
                )

            # Create splines
            self.throughput_interp = make_interp_spline(sep, throughput, k=3)
            self.raw_contrast_interp = make_interp_spline(sep, raw_contrast, k=3)

            # Compute the Inner Working Angle
            half_max_throughput = max(throughput) / 2
            closest_ind = np.searchsorted(throughput, half_max_throughput)

            def iwa_func(x):
                return self.throughput_interp(x) - half_max_throughput

            self.IWA = (
                root_scalar(
                    iwa_func, bracket=[sep[closest_ind - 1], sep[closest_ind]]
                ).root
                * lod
            )
        else:
            logger.warning("2d contrast/throughput not supported currently")

        logger.info(f"Created {yip_path.stem}")

    def create_psf_datacube(self, batch_size=128):
        """Load the PSF datacube from a file or generate it if it doesn't exist.

        The PSF datacube is a 4D array of PSFs at each pixel (x psf offset,
        y psf offset, x, y). Given the computational cost of generating this
        datacube, it is only generated when needed and saved to a numpy binary
        file in the yip_path directory.

        Args:
            batch_size (int):
                Number of PSFs to generate in each batch. Default is 128.
        """
        datacube_path = Path(self.yip_path, "psf_datacube.npy")
        if datacube_path.exists():
            logger.info(f"Loading PSF datacube from {datacube_path}.")
            psfs = jnp.load(datacube_path)
        else:
            # Create data cube of spatially dependent PSFs.
            psfs_shape = (*self.psf_shape, *self.psf_shape)
            psfs = np.zeros(psfs_shape, dtype=np.float32)
            pixel_lod = (
                (np.arange(self.npixels) - ((self.npixels - 1) // 2))
                * u.pixel
                * self.pixel_scale
            ).value

            # Get the pixel coordinates for the PSF evaluations
            x_lod, y_lod = np.meshgrid(pixel_lod, pixel_lod, indexing="xy")
            points = np.column_stack((x_lod.flatten(), y_lod.flatten()))
            n_points = points.shape[0]

            logger.info(
                "Calculating data cube of spatially dependent PSFs, please hold..."
            )
            with tqdm(total=n_points, desc="Computing PSFs") as pb:
                for i in range(0, n_points, batch_size):
                    # Select the current batch
                    batch_points = points[i : i + batch_size]
                    batch_psfs = self.offax.create_psfs_parallel(
                        batch_points[:, 0], batch_points[:, 1]
                    )

                    # Store the batch in the data cube
                    psfs.reshape((-1, self.npixels, self.npixels))[
                        i : i + batch_size
                    ] = batch_psfs
                    pb.update(batch_points.shape[0])
            jnp.save(datacube_path, psfs)
            logger.info(f"PSF datacube saved to {datacube_path}.")

        self.has_psf_datacube = True
        self.psf_datacube = psfs

    def __repr__(self):
        """String representation of the Coronagraph object."""
        base_str = f"Coronagraph {self.name} ({self.yip_path})\n"
        base_str += f"{self.offax.type} off-axis PSFs, {self.offax.n_psfs} provided"
        if self.has_psf_datacube:
            base_str += f"\n{base_str}\nPSF datacube loaded"
        return base_str

    def get_throughput_curve(self, aperture_radius_lod=0.7, oversample=2, plot=True):
        """Creates the coronagraph throughput curve.

        Compute the coronagraph throughput vs. separation using ONLY the
        provided planet off-axis PSFs (no interpolation).
        We define throughput as the fraction of the total flux
        (planet PSF normalized to 1) that lands inside a photometric aperture.
        """
        separations = []
        throughputs = []

        # Loop through all provided offsets
        for i, x_lod_val in enumerate(np.array(self.offax.x_offsets)):
            for j, y_lod_val in enumerate(np.array(self.offax.y_offsets)):
                psf_img = self.offax.reshaped_psfs[i, j]
                r = np.sqrt(x_lod_val**2 + y_lod_val**2)

                # Aperture radius in pixel units
                radius_pix = aperture_radius_lod / self.pixel_scale.value

                # Planet coords in pixel space
                px = convert_to_pix(
                    x_lod_val, self.offax.center_x, self.pixel_scale
                ).value.astype(int)
                py = convert_to_pix(
                    y_lod_val, self.offax.center_y, self.pixel_scale
                ).value.astype(int)

                # Extract & oversample
                subarr_oversamp, px_os, py_os, radius_os, subarr_orig = (
                    extract_and_oversample_subarray(
                        psf_img, px, py, radius_pix, oversample
                    )
                )

                # Measure flux in aperture
                planet_flux_in_ap = measure_flux_in_oversampled_aperture(
                    subarr_oversamp, px_os, py_os, radius_os, subarr_orig
                )

                # planet_flux_in_ap is the throughput if psf_img sums to 1
                throughput = planet_flux_in_ap
                separations.append(r)
                throughputs.append(throughput)

        # Sort by separation
        separations = np.array(separations)
        throughputs = np.array(throughputs)
        idx_sort = np.argsort(separations)
        separations = separations[idx_sort]
        throughputs = throughputs[idx_sort]

        if plot:
            import matplotlib.pyplot as plt

            plt.figure()
            plt.plot(separations, throughputs, "o-", ms=6)
            plt.xlabel("Separation [λ/D]")
            plt.ylabel("Throughput")
            plt.title(f"{self.name} Throughput")
            plt.grid(True)
            plt.show()

        return separations, throughputs

    def get_contrast_curve(
        self, stellar_diam=0 * lod, aperture_radius_lod=0.7, oversample=2, plot=True
    ):
        """Creates the raw contrast curve.

        Compute a photometric aperture–based contrast curve vs. separation,
        using the provided offsets.

        Args:
            stellar_diam (Quantity):
                The stellar diameter used for the star's PSF.
            aperture_radius_lod (float):
                The aperture radius in lambda/D.
            oversample (int):
                The oversampling factor for interpolation.
            plot (bool):
                Whether to plot the contrast curve.
        """
        # Grab star PSF for given stellar diameter
        star_psf = self.stellar_intens(stellar_diam)

        separations = []
        contrasts = []

        for i, x_lod_val in enumerate(np.array(self.offax.x_offsets)):
            for j, y_lod_val in enumerate(np.array(self.offax.y_offsets)):
                # Planet PSF & separation
                planet_psf = self.offax.reshaped_psfs[i, j]
                r = np.sqrt(x_lod_val**2 + y_lod_val**2)

                # Aperture radius in pixel units
                radius_pix = aperture_radius_lod / self.pixel_scale.value

                # Planet flux in aperture
                px = convert_to_pix(
                    x_lod_val, self.offax.center_x, self.pixel_scale
                ).value.astype(int)
                py = convert_to_pix(
                    y_lod_val, self.offax.center_y, self.pixel_scale
                ).value.astype(int)
                subarr_oversamp_p, px_os, py_os, radius_os, subarr_orig_p = (
                    extract_and_oversample_subarray(
                        planet_psf, px, py, radius_pix, oversample
                    )
                )
                planet_flux_in_ap = measure_flux_in_oversampled_aperture(
                    subarr_oversamp_p, px_os, py_os, radius_os, subarr_orig_p
                )

                # Star flux in aperture at same offset
                subarr_oversamp_s, sx_os, sy_os, radius_os_s, subarr_orig_s = (
                    extract_and_oversample_subarray(
                        star_psf, px, py, radius_pix, oversample
                    )
                )
                star_flux_in_ap = measure_flux_in_oversampled_aperture(
                    subarr_oversamp_s, sx_os, sy_os, radius_os_s, subarr_orig_s
                )

                # Contrast
                if star_flux_in_ap > 0:
                    contrast_val = star_flux_in_ap / planet_flux_in_ap
                else:
                    contrast_val = 0

                separations.append(r)
                contrasts.append(contrast_val)

        # Sort by separation
        separations = np.array(separations)
        contrasts = np.array(contrasts)
        idx_sort = np.argsort(separations)
        separations = separations[idx_sort]
        contrasts = contrasts[idx_sort]

        if plot:
            import matplotlib.pyplot as plt

            plt.figure()
            plt.plot(separations, contrasts, "o-")
            plt.xlabel("Separation [λ/D]")
            plt.ylabel("Raw Contrast")
            plt.title(f"{self.name} Raw Contrast")
            plt.yscale("log")
            plt.grid(True)
            plt.show()

        return separations, contrasts



"""Logging module."""

import logging

lib_name = "yippy"
# See https://gist.github.com/fnky/458719343aabd01cfb17a3a4f7296797 for
# info on the color codes
lib_color = "229"


# ANSI escape sequences for colors
class ColorCodes:
    """ANSI escape sequences for colors."""

    RESET = "\033[0m"
    BLACK = "\033[30m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    LIB = f"\033[38;5;{lib_color}m"


# Custom formatter to add colors
class ColorFormatter(logging.Formatter):
    """Custom formatter to add colors to log messages."""

    COLORS = {
        logging.DEBUG: ColorCodes.BLUE,
        logging.INFO: ColorCodes.GREEN,
        logging.WARNING: ColorCodes.YELLOW,
        logging.ERROR: ColorCodes.RED,
        logging.CRITICAL: ColorCodes.MAGENTA,
    }

    def format(self, record: logging.LogRecord):
        """Format the log message with colors."""
        log = super().format(record)
        color = self.COLORS.get(record.levelno, ColorCodes.WHITE)
        return f"{ColorCodes.LIB}\033[48;5;16m[{lib_name}]\033[0m {color}{log}"


logger = logging.getLogger(f"{lib_name}")

shell_handler = logging.StreamHandler()
file_handler = logging.FileHandler("debug.log")

logger.setLevel(logging.DEBUG)
shell_handler.setLevel(logging.INFO)
file_handler.setLevel(logging.DEBUG)

shell_fmt = "%(levelname)s [%(asctime)s] \033[0m%(message)s"
file_fmt = (
    f"[{lib_name}] %(levelname)s %(asctime)s [%(filename)s:"
    "%(funcName)s:%(lineno)d] %(message)s"
)
shell_formatter = ColorFormatter(shell_fmt)
file_formatter = logging.Formatter(file_fmt)

shell_handler.setFormatter(shell_formatter)
file_handler.setFormatter(file_formatter)

logger.addHandler(shell_handler)
logger.addHandler(file_handler)

logger.propagate = True
